#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyradium.xmlhooks.XMLHookRegistry import BaseHook, XMLHookRegistry
from pyradium.Exceptions import NoAgendaException

@XMLHookRegistry.register_hook
class AgendaHook(BaseHook):
	_TAG_NAME = "agenda"

	@classmethod
	def handle(cls, rendered_presentation, node):
		meta = rendered_presentation.renderer.presentation.meta
		agenda = meta.get("agenda")
		if agenda is None:
			raise NoAgendaException("s:agenda requested but no agenda defined in metadata.")
		replacement_node = node.ownerDocument.createElement("ul")
		for item in agenda:
			li_node = replacement_node.appendChild(node.ownerDocument.createElement("li"))
			li_node.appendChild(node.ownerDocument.createElement("b")).appendChild(node.ownerDocument.createTextNode(f"{item.start_time} - {item.end_time}"))
			li_node.appendChild(node.ownerDocument.createTextNode(f": {item.text}"))
		replacement_node.setAttribute("class", "agenda")
		return replacement_node



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import logging
import subprocess
import datetime
from .Exceptions import DeploymentException
from .CmdlineEscape import CmdlineEscape

_log = logging.getLogger(__spec__.name)

class Deployment():
	def __init__(self, deployment_configuration: dict, presentation_directory: str):
		self._deployment_configuration = deployment_configuration
		self._presentation_directory = presentation_directory

	def deploy(self, configuration_name: str):
		_log.debug("Executing deployment configuration: %s", configuration_name)
		if configuration_name not in self._deployment_configuration:
			raise DeploymentException(f"No such deployment configuration: {configuration_name}")

		target_config = self._deployment_configuration[configuration_name]
		if "available" in target_config:
			try:
				ts = datetime.datetime.strptime(target_config["available"], "%Y-%m-%d %H:%M:%S")
			except ValueError as e:
				raise DeploymentException(f"Deployment configuration {configuration_name} has invalid format for 'available' datetime: {target_config['available']}. Expecting format \"YYYY-mm-dd HH:MM:SS\".") from e
			if "not_available_redirect" in target_config:
				redirect = target_config["not_available_redirect"]
			else:
				redirect = "/not_available_yet.html"
			with open(self._presentation_directory + "/.htaccess", "w") as f:
				if "access_key" in target_config:
					print("# If access key is set in query string, set the access_key cookie", file = f)
					print(f"RewriteCond %{{QUERY_STRING}} ^{target_config['access_key']}$", file = f)
					print(f"RewriteRule ^ - [CO=access_key:{target_config['access_key']}:%{{HTTP_HOST}}] [L]", file = f)
					print(file = f)

				if "access_key" not in target_config:
					print("# If too soon, redirect to 'not available' page", file = f)
				else:
					print("# If too soon, access_key query string is not set and access_key cookie is not set, redirect to 'not available' page", file = f)
				print(f"RewriteCond %{{TIME}} <{ts.strftime('%Y%m%d%H%M%S')}", file = f)
				if "access_key" in target_config:
					print(f"RewriteCond %{{QUERY_STRING}} !^{target_config['access_key']}$", file = f)
					print(f"RewriteCond %{{HTTP_COOKIE}} !access_key={target_config['access_key']}", file = f)
				print(f"RewriteRule ^ {redirect} [R=307]", file = f)

		cmd = [ "rsync", "-a", "--mkpath", "--delete", "--chmod=D755,F644", self._presentation_directory + "/", target_config["remote_target"] ]
		_log.debug("Synchronizing via rsync: %s", CmdlineEscape().cmdline(cmd))
		try:
			subprocess.check_call(cmd)
		except subprocess.CalledProcessError as e:
			raise DeploymentException("Deployment {configuration_name} failed to successfully execute rsync: {CmdlineEscape().cmdline(cmd)}") from e
		_log.info("Successully deployed deployment configuration \"%s\".", configuration_name)

	def deploy_all(self):
		for configuration_name in self._deployment_configuration:
			self.deploy(configuration_name)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyradium.xmlhooks.XMLHookRegistry import BaseHook, XMLHookRegistry
from pyradium.Tools import XMLTools
from pyradium.Enums import PresentationFeature

@XMLHookRegistry.register_hook
class AcronymHook(BaseHook):
	_TAG_NAME = "ac"

	@classmethod
	def handle(cls, rendered_presentation, node):
		acronym_id = XMLTools.inner_text(node)

		acronym = rendered_presentation.renderer.get_custom_renderer("acronym")
		resolved = acronym.resolve(acronym_id)
		if resolved is None:
			replacement_node = node.ownerDocument.createTextNode(acronym_id)
		else:
			rendered_presentation.add_feature(PresentationFeature.Acronyms)

			outer_span_node = node.ownerDocument.createElement("span")
			replacement_node = outer_span_node

			if resolved.uri is not None:
				link_node = outer_span_node.appendChild(node.ownerDocument.createElement("a"))
				link_node.setAttribute("href", resolved.uri)
				link_node.setAttribute("class", "tooltip-link")
				chain_node = link_node
			else:
				chain_node = outer_span_node

			chain_node.setAttribute("class", "tooltip")
			chain_node.appendChild(node.ownerDocument.createTextNode(resolved.acronym))

			inner_span = chain_node.appendChild(node.ownerDocument.createElement("span"))
			inner_span.setAttribute("class", "text")
			inner_span.appendChild(node.ownerDocument.createTextNode(resolved.text))

		return replacement_node



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import logging
from .BaseAction import BaseAction
from .RenderingParameters import RenderingParameters
from .Presentation import Presentation
from .OrderedSet import OrderedSet
from .Slide import RenderSlideDirective
from .Acronyms import AcronymDirective
from .Tools import XMLTools, HashTools
from .CmdlineParser import CmdlineParser

_log = logging.getLogger(__spec__.name)

class ActionHashPresentation(BaseAction):
	def run(self):
		self._rendering_parameters = RenderingParameters(include_dirs = [ os.path.dirname(self._args.infile) or "." ] + self._args.include_dir)
		presentation = Presentation.load_from_file(self._args.infile, self._rendering_parameters)

		# The first dependencies are the XML files themselves
		self._dependencies = OrderedSet()
		self._dependencies |= presentation.sources

		# Then we iterate throught the directives
		for directive in presentation.content:
			self._handle_directive(directive)

		hash_value = HashTools.hash_files(self._dependencies)
		print(hash_value)

	def _add_dependency(self, relative_filename):
		abs_filename = self._rendering_parameters.include_dirs.lookup(relative_filename)
		self._dependencies.add(abs_filename)

	def _visit_node(self, node):
		if node.tagName == "s:img":
			self._add_dependency(node.getAttribute("src"))
		elif node.tagName in [ "s:code", "s:term" ]:
			if node.hasAttribute("src"):
				self._add_dependency(node.getAttribute("src"))
		elif node.tagName == "s:exec":
			cmdline = CmdlineParser().parse(node.getAttribute("cmd"))
			self._add_dependency(cmdline[0])

	def _handle_directive(self, directive):
		if isinstance(directive, RenderSlideDirective):
			XMLTools.walk_elements(directive.xmlnode, self._visit_node)
		elif isinstance(directive, AcronymDirective):
			self._add_dependency(directive.src)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from .BaseAction import BaseAction

_KNOWN_TEMPLATES = {
	"slide_title": """
	<slide type="title"/>
	""",

	"slide_toc": """
	<slide type="toc"/>
	""",

	"slide_sectiontitle": """
	<slide type="sectiontitle">
		<s:time abs="10 sec" />
	</slide>
	""",

	"slide_leftright": """
	<slide type="leftright">
	<s:var name="heading" value="" />
		<s:content name="left">
			Left
		</s:content>
		<s:content name="right">
			Right
		</s:content>
	</slide>
	""",

	"slide_quote": """
	<slide type="quote">
		<s:time abs="15 sec"/>
		Quote
		<s:var name="author" value="Author"/>
	</slide>
	""",

	"slide_final": """
	<slide type="final">
		<s:time rel="0" />
	</slide>
	""",

	"slide_acronyms": """
	<slide type="acronyms">
		<s:time rel="0" />
	</slide>
	""",

	"slide_feedback": """
	<slide type="feedback">
		<s:time rel="0" />
		<s:var name="endpoint" value="https://my_server.com/feedback/submit" />
	</slide>
	""",

	"textblock_code": """
		<s:code lang="X"><![CDATA[
		]]></s:code>
	""",

	"textblock_term": """
		<s:term prompt="$ "><![CDATA[
		]]></s:term>
	""",


	"image_img": """<s:img src="X" />""",
	"image_anim": """
	<slide type="animation">
		<s:var name="heading" value="X" />
		<s:var name="filename" value="X.svg" />
	</slide>
	""",
	"image_plot": """<s:plot src="X" />""",
	"image_graphviz": """<s:graphviz src="X" />""",

	"other_circuitjs": """
	<slide>
		<s:var name="heading" value="Title"/>
		<s:circuit>
			<s:param name="name" value="X"/>
			<s:param name="content" value="image"/>
			<s:param name="src" value="*"/>
		</s:circuit>
	</slide>
	""",
}

class ActionTemplateHelper(BaseAction):
	def run(self):
		if self._args.template_name not in _KNOWN_TEMPLATES:
			raise Exception("No such template: %s (must be one of %s)" % (self._args.template_name, ", ".join(sorted(_KNOWN_TEMPLATES))))
		text = _KNOWN_TEMPLATES[self._args.template_name]
		text = text.lstrip("\n").rstrip("\r\n\t ")
		print(text)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import contextlib
from .SVGStyle import SVGStyle
from .Vector2D import Vector2D

class SVGPath():
	def __init__(self, path_node):
		self._node = path_node
		self._style = SVGStyle.from_node(self._node, auto_sync = True).default_path()
		self._cmds = [ ]
		self._pos = Vector2D()

	@property
	def pos(self):
		return self._pos

	@property
	def style(self):
		return self._style

	@contextlib.contextmanager
	def returnto(self):
		prev = self._pos
		yield
		self.move_to(prev.x, prev.y)

	def _append(self, *cmds):
		self._cmds += cmds
		self._node.setAttribute("d", " ".join(self._cmds))

	def move_to(self, x, y):
		self._append("M", f"{x},{y}")
		self._pos = Vector2D(x, y)
		return self

	def move_rel(self, x, y):
		self._append("m", f"{x},{y}")
		self._pos += Vector2D(x, y)
		return self

	def line_to(self, x, y):
		self._append("L", f"{x},{y}")
		self._pos += Vector2D(x, y)
		return self

	def line_rel(self, x, y):
		self._append("l", f"{x},{y}")
		self._pos += Vector2D(x, y)
		return self

	def horiz_rel(self, x):
		self._append("h", f"{x}")
		self._pos += Vector2D(x, 0)
		return self

	def vert_rel(self, y):
		self._append("v", f"{y}")
		self._pos += Vector2D(0, y)
		return self



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from .SVGStyle import SVGStyle

class SVGLayer():
	def __init__(self, group_node):
		self._node = group_node

	@property
	def label(self):
		return self._node.getAttribute("inkscape:label")

	@property
	def layer_id(self):
		return self._node.getAttribute("id")

	@property
	def is_visible(self):
		return SVGStyle.from_node(self._node).is_visible

	def modify_style(self, callback):
		style = SVGStyle.from_node(self._node)
		callback(style)
		style.sync_to_node()

	def hide(self):
		def _callback(style):
			style["display"] = "none"
		self.modify_style(_callback)

	def show(self):
		def _callback(style):
			style["display"] = "inline"
		self.modify_style(_callback)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import urllib.parse
import lzstr
from pyradium.Tools import XMLTools
from pyradium.Exceptions import MissingParameterException

class CircuitJSCircuit():
	_DEFAULTS = {
		"euroResistors":	"true",
		"IECGates":			"false",
		"whiteBackground":	"true",
		"positiveColor":	"#27ae60",
		"negativeColor":	"#c0392b",
		"selectColor":		"#2c3e50",
	}
	_DEFAULT_URI = "https://www.falstad.com/circuit/circuitjs.html"
	_RESET_CONFIG = "$ 0 5e-6 10.2 50 5 50 5e-11"

	def __init__(self, circuit_text: str | None = None, circuit_params: dict | None = None, uri: str = _DEFAULT_URI, presentation_params: dict | None = None, display_content: list | None = None, original_xml_node = None):
		self._circuit = None
		self.circuit_text = circuit_text
		self._circuit_params = circuit_params if (circuit_params is not None) else { }
		self._uri = uri
		self._presentation_params = presentation_params if (presentation_params is not None) else { }
		self._display_content = display_content if (display_content is not None) else [ ]
		self._original_xml_node = original_xml_node

	def reset_presentation_parameters(self):
		if self._circuit is None:
			self._circuit = [ self._RESET_CONFIG ]
		else:
			self._circuit[0] = self._RESET_CONFIG

	@property
	def circuit_text(self):
		if self._circuit is None:
			return None
		else:
			return "\n".join(self._circuit) + "\n"

	@classmethod
	def automatic_filename_from_circuit_name(cls, circuit_name):
		return f"circuit_{circuit_name}.txt"

	@property
	def automatic_filename(self):
		return self.automatic_filename_from_circuit_name(self.get_presentation_parameter("name"))

	def _parse_text(self, text):
		text = text.strip("\r\n")
		text = [ line.strip(" \t\r\n") for line in text.split("\n") ]
		return text

	@circuit_text.setter
	def circuit_text(self, value):
		if value is None:
			self._circuit = None
		else:
			self._circuit = self._parse_text(value)

	@property
	def display_content(self):
		return self._display_content

	def circuit_params(self, include_circuit: bool = False):
		args = dict(self._circuit_params)
		if include_circuit and (self._circuit is not None):
			ctz = lzstr.LZStringCompressor.compress_to_url_component(self.circuit_text.encode("ascii"), escape = False)
			args["ctz"]	= ctz
		return args

	def full_url(self, include_circuit: bool = False):
		args = self.circuit_params(include_circuit = include_circuit)
		if len(args) == 0:
			return self._uri
		else:
			return f"{self._uri}?{urllib.parse.urlencode(args)}"

	def get_presentation_parameter(self, name):
		return self._presentation_params.get(name)

	def has_presentation_parameter(self, name):
		return name in self._presentation_params

	@classmethod
	def from_xml(cls, xml_node, no_defaults: bool = False, find_file_function = None):
		circuit_text = None
		if no_defaults:
			circuit_params = { }
		else:
			circuit_params = dict(cls._DEFAULTS)
		uri = cls._DEFAULT_URI
		presentation_params = { }
		display_content = [ ]

		for child_node in xml_node.childNodes:
			if (child_node.nodeType == child_node.ELEMENT_NODE) and (child_node.nodeName == "s:param"):
				# Parse parameter
				name = child_node.getAttribute("name")
				value = XMLTools.get_node_value(child_node, find_file_function = find_file_function).strip()
				if name == "uri":
					uri = value
				elif name == "src":
					circuit_text = value
				elif name == "srclink":
					parsed_url = urllib.parse.urlparse(value)
					query = urllib.parse.parse_qs(parsed_url.query)
					if "ctz" not in query:
						raise MissingParameterException(f"The URI provided as a 'srclink' for the 's:circuit' tag is missing the ctz= portion in its query string: {value}")
					srclink = query["ctz"][0]
					circuit_text = lzstr.LZStringDecompressor.decompress_from_url_component(srclink, escape = False).decode("ascii")
				elif name in [ "name", "content" ]:
					presentation_params[name] = value
				else:
					circuit_params[name] = value
			else:
				display_content.append(child_node)

			if circuit_text == "*":
				# Automatically load from filename
				if "name" not in presentation_params:
					raise MissingParameterException("Automatic determination of circuit source requested by using *, but circuit has no 'name' parameter set.")
				if find_file_function is None:
					raise MissingParameterException("Automatic determination of circuit source requested by using *, but no way of determining the included files given.")
				with open(find_file_function(cls.automatic_filename_from_circuit_name(presentation_params["name"]))) as f:
					circuit_text = f.read()
		return cls(circuit_text = circuit_text, circuit_params = circuit_params, uri = uri, presentation_params = presentation_params, display_content = display_content, original_xml_node = xml_node)

	def _find_circuit_nodes(self):
		found = [ ]
		for node in XMLTools.findall(self._original_xml_node, "s:param"):
			name = node.getAttribute("name")
			if name in [ "src", "srclink" ]:
				found.append(node)
		return found

	def _replace_source_node(self, replacement):
		present_nodes = self._find_circuit_nodes()
		first_node = present_nodes[0] if (len(present_nodes) > 0) else None

		if (len(present_nodes) == 1) and (replacement.getAttribute("src") == first_node.getAttribute("src")):
			current_text = XMLTools.inner_text(first_node)
			replacement_text = XMLTools.inner_text(replacement)
			if self._parse_text(current_text) == self._parse_text(replacement_text):
				# Nothing needs to change.
				return False

		if len(present_nodes) > 0:
			XMLTools.replace_node(first_node, replacement)
			for node in present_nodes[1:]:
				XMLTools.remove_node(node)
		else:
			self._original_xml_node.appendChild(replacement)
		return True

	def modify_dom_source_inline(self):
		replacement = self._original_xml_node.ownerDocument.createElement("s:param")
		replacement.setAttribute("name", "src")
		indented_text = "\n\t\t\t\t" + self.circuit_text.rstrip("\n").replace("\n", "\n\t\t\t\t") + "\n\t\t\t"
		replacement.appendChild(self._original_xml_node.ownerDocument.createTextNode(indented_text))
		return self._replace_source_node(replacement)

	def modify_dom_source_external_file(self, filename):
		replacement = self._original_xml_node.ownerDocument.createElement("s:param")
		replacement.setAttribute("name", "src")
		if filename != "*":
			replacement.setAttribute("src", filename)
		else:
			replacement.setAttribute("value", "*")
		return self._replace_source_node(replacement)

	def __str__(self):
		args = [ ]
		if self._circuit is None:
			args.append("no data")
		else:
			args.append(f"{len(self._circuit)} lines")
		args.append(self.full_url(True))
		return f"CircuitJS<{', '.join(args)}>"



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2021-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import dataclasses
import os
from pyradium.svg import SVGValidator, SVGValidatorErrorClass
from .FileLookup import FileLookup

@dataclasses.dataclass()
class RenderingParameters():
	template_style: str = "default"
	template_style_opts: list[str] | None = None
	honor_pauses: bool = True
	collapse_animation: bool = False
	template_dirs: list[str] | None = dataclasses.field(default = None, init = False)
	extra_template_dirs: list[str] | None = None
	include_dirs: list[str] | None = None
	index_filename: str = "index.html"
	resource_uri: str = ""
	geometry: tuple = (1280, 720)
	image_max_dimension: int = 1920
	presentation_features: set | None = None
	injected_metadata: dict | None = None
	trustworthy_source: bool = False
	allow_missing_svg_fonts: bool = False
	svg_validator: None = dataclasses.field(default = None, init = False)

	@property
	def geometry_x(self):
		return self.geometry[0]

	@property
	def geometry_y(self):
		return self.geometry[1]

	def __post_init__(self):
		# Initialize default values
		if self.template_style_opts is None:
			self.template_style_opts = [ ]
		if self.extra_template_dirs is None:
			self.extra_template_dirs = [ ]
		if self.include_dirs is None:
			self.include_dirs = [ ]
		if self.presentation_features is None:
			self.presentation_features = frozenset()
		else:
			self.presentation_features = frozenset(self.presentation_features)

		# Compute non-settable variables and initialize wrappers
		self.template_dirs = FileLookup([ os.path.expanduser("~/.config/pyradium/templates"), os.path.dirname(os.path.realpath(__file__)) + "/templates" ] + self.extra_template_dirs)
		self.include_dirs = FileLookup(self.include_dirs)

		# Initialize SVG validator
		if self.allow_missing_svg_fonts:
			self.svg_validator = SVGValidator(check_missing_fonts = SVGValidatorErrorClass.EmitWarning)
		else:
			self.svg_validator = SVGValidator(check_missing_fonts = SVGValidatorErrorClass.ThrowException)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import unittest
from pyradium.Agenda import Agenda
from pyradium.Exceptions import IllegalAgendaSyntaxException, UndefinedAgendaTimeException, UnresolvableWeightedEntryException, AgendaTimeMismatchException

class AgendaTests(unittest.TestCase):
	def test_simple1(self):
		agenda = Agenda.parse("""
		14:00
		+1:00	Presentation
		""")
		self.assertEqual(len(agenda), 1)
		self.assertEqual(agenda[0].start_time, "14:00")
		self.assertEqual(agenda[0].end_time, "15:00")
		self.assertEqual(agenda[0].text, "Presentation")

	def test_simple2(self):
		agenda = Agenda.parse("""
		14:00
		+1:00	Presentation
		+0:45	XYZ
		+0:45	ABC
		""")
		self.assertEqual(len(agenda), 3)
		self.assertEqual(agenda[0].start_time, "14:00")
		self.assertEqual(agenda[0].end_time, "15:00")
		self.assertEqual(agenda[0].text, "Presentation")
		self.assertEqual(agenda[1].start_time, "15:00")
		self.assertEqual(agenda[1].end_time, "15:45")
		self.assertEqual(agenda[1].text, "XYZ")
		self.assertEqual(agenda[2].start_time, "15:45")
		self.assertEqual(agenda[2].end_time, "16:30")
		self.assertEqual(agenda[2].text, "ABC")

	def test_variable_simple3(self):
		agenda = Agenda.parse("""
		13:00
		14:00	A
		15:15	B
		16:20	C
		""")
		self.assertEqual(len(agenda), 3)
		self.assertEqual(agenda[0].start_time, "13:00")
		self.assertEqual(agenda[0].end_time, "14:00")
		self.assertEqual(agenda[1].start_time, "14:00")
		self.assertEqual(agenda[1].end_time, "15:15")
		self.assertEqual(agenda[2].start_time, "15:15")
		self.assertEqual(agenda[2].end_time, "16:20")

	def test_marker(self):
		agenda = Agenda.parse("""
		14:00
		:Beg
		+1:00	Presentation
		:First
		+0:45	XYZ
		+0:45	ABC
		:End
		""")
		self.assertEqual(len(agenda), 3)
		self.assertEqual(agenda[0].start_time, "14:00")
		self.assertEqual(agenda[0].end_time, "15:00")
		self.assertEqual(agenda[0].text, "Presentation")
		self.assertEqual(agenda[1].start_time, "15:00")
		self.assertEqual(agenda[1].end_time, "15:45")
		self.assertEqual(agenda[1].text, "XYZ")
		self.assertEqual(agenda[2].start_time, "15:45")
		self.assertEqual(agenda[2].end_time, "16:30")
		self.assertEqual(agenda[2].text, "ABC")
		self.assertEqual(list(agenda.markers), [ ("Beg", "14:00"), ("First", "15:00"), ("End", "16:30") ])

	def test_backwards1(self):
		agenda = Agenda.parse("""
		+0:15	A
		+0:30	B
		+1:00	C
		14:00
		""")
		self.assertEqual(len(agenda), 3)
		self.assertEqual(agenda[0].start_time, "12:15")
		self.assertEqual(agenda[0].end_time, "12:30")
		self.assertEqual(agenda[1].start_time, "12:30")
		self.assertEqual(agenda[1].end_time, "13:00")
		self.assertEqual(agenda[2].start_time, "13:00")
		self.assertEqual(agenda[2].end_time, "14:00")

	def test_backwards2(self):
		agenda = Agenda.parse("""
		+0:15	A
		+0:30	B
		+1:00	C
		14:00
		*		D
		14:15
		""")
		self.assertEqual(len(agenda), 4)
		self.assertEqual(agenda[0].start_time, "12:15")
		self.assertEqual(agenda[0].end_time, "12:30")
		self.assertEqual(agenda[1].start_time, "12:30")
		self.assertEqual(agenda[1].end_time, "13:00")
		self.assertEqual(agenda[2].start_time, "13:00")
		self.assertEqual(agenda[2].end_time, "14:00")
		self.assertEqual(agenda[3].start_time, "14:00")
		self.assertEqual(agenda[3].end_time, "14:15")

	def test_variable_gap(self):
		agenda = Agenda.parse("""
		13:00
		14:00	A
		15:00
		16:00	B
		""")
		self.assertEqual(len(agenda), 2)
		self.assertEqual(agenda[0].start_time, "13:00")
		self.assertEqual(agenda[0].end_time, "14:00")
		self.assertEqual(agenda[1].start_time, "15:00")
		self.assertEqual(agenda[1].end_time, "16:00")

	def test_past_midnight1(self):
		agenda = Agenda.parse("""
		23:00
		+2:00	Presentation
		""")
		self.assertEqual(len(agenda), 1)
		self.assertEqual(agenda[0].start_time, "23:00")
		self.assertEqual(agenda[0].end_time, "1:00")
		self.assertEqual(agenda[0].text, "Presentation")

	def test_past_midnight2(self):
		agenda = Agenda.parse("""
		23:00
		1+1:00	A
		3:00	B
		""")
		self.assertEqual(len(agenda), 2)
		self.assertEqual(agenda[0].start_time, "23:00")
		self.assertEqual(agenda[0].end_time, "1:00")
		self.assertEqual(agenda[0].duration, "2:00")
		self.assertEqual(agenda[0].text, "A")
		self.assertEqual(agenda[1].start_time, "1:00")
		self.assertEqual(agenda[1].end_time, "3:00")
		self.assertEqual(agenda[1].duration, "2:00")
		self.assertEqual(agenda[1].text, "B")

	def test_variable_simple1(self):
		agenda = Agenda.parse("""
		13:00
		*		A
		14:00
		""")
		self.assertEqual(len(agenda), 1)
		self.assertEqual(agenda[0].start_time, "13:00")
		self.assertEqual(agenda[0].end_time, "14:00")

	def test_variable_simple2(self):
		agenda = Agenda.parse("""
		13:00
		*		A
		*		B
		14:00
		""")
		self.assertEqual(len(agenda), 2)
		self.assertEqual(agenda[0].start_time, "13:00")
		self.assertEqual(agenda[0].end_time, "13:30")
		self.assertEqual(agenda[1].start_time, "13:30")
		self.assertEqual(agenda[1].end_time, "14:00")

	def test_variable_complex1(self):
		agenda = Agenda.parse("""
		13:00
		*		A
		+0:30	B
		*		C
		14:00
		""")
		self.assertEqual(len(agenda), 3)
		self.assertEqual(agenda[0].start_time, "13:00")
		self.assertEqual(agenda[0].end_time, "13:15")
		self.assertEqual(agenda[1].start_time, "13:15")
		self.assertEqual(agenda[1].end_time, "13:45")
		self.assertEqual(agenda[2].start_time, "13:45")
		self.assertEqual(agenda[2].end_time, "14:00")

	def test_variable_complex2(self):
		agenda = Agenda.parse("""
		13:00
		*		A
		+0:30	B
		*2		C
		14:00
		""")
		self.assertEqual(len(agenda), 3)
		self.assertEqual(agenda[0].start_time, "13:00")
		self.assertEqual(agenda[0].end_time, "13:10")
		self.assertEqual(agenda[1].start_time, "13:10")
		self.assertEqual(agenda[1].end_time, "13:40")
		self.assertEqual(agenda[2].start_time, "13:40")
		self.assertEqual(agenda[2].end_time, "14:00")

	def test_variable_complex3(self):
		agenda = Agenda.parse("""
		13:00
		*		A
		+0:30	B
		*2		C
		+0:10	D
		*		E
		14:00
		""")
		self.assertEqual(len(agenda), 5)
		self.assertEqual(agenda[0].start_time, "13:00")
		self.assertEqual(agenda[0].end_time, "13:05")
		self.assertEqual(agenda[1].start_time, "13:05")
		self.assertEqual(agenda[1].end_time, "13:35")
		self.assertEqual(agenda[2].start_time, "13:35")
		self.assertEqual(agenda[2].end_time, "13:45")
		self.assertEqual(agenda[3].start_time, "13:45")
		self.assertEqual(agenda[3].end_time, "13:55")
		self.assertEqual(agenda[4].start_time, "13:55")
		self.assertEqual(agenda[4].end_time, "14:00")

	def test_rounding(self):
		agenda = Agenda.parse("""
		13:01
		14:02	A
		15:03	B
		""", granularity_minutes = 5)
		self.assertEqual(len(agenda), 2)
		self.assertEqual(agenda[0].start_time, "13:00")
		self.assertEqual(agenda[0].end_time, "14:00")
		self.assertEqual(agenda[1].start_time, "14:00")
		self.assertEqual(agenda[1].end_time, "15:05")

	def test_divisor(self):
		agenda = Agenda.parse("""
		14:00
		+1:00/2		A
		+1:00/4		B
		""")
		self.assertEqual(len(agenda), 2)
		self.assertEqual(agenda[0].start_time, "14:00")
		self.assertEqual(agenda[0].end_time, "14:30")
		self.assertEqual(agenda[1].start_time, "14:30")
		self.assertEqual(agenda[1].end_time, "14:45")

	def test_error_syntax1(self):
		with self.assertRaises(IllegalAgendaSyntaxException):
			Agenda.parse("foo")

	def test_error_syntax2(self):
		with self.assertRaises(IllegalAgendaSyntaxException):
			Agenda.parse("1:00/2	A")

	def test_error_underspecified(self):
		with self.assertRaises(UndefinedAgendaTimeException):
			Agenda.parse("+1:00	A")

	def test_error_weights_zero(self):
		with self.assertRaises(UnresolvableWeightedEntryException):
			Agenda.parse("""
			13:00
			*0	foo
			14:00
			""")

	def test_error_weighted_out_of_time(self):
		with self.assertRaises(UnresolvableWeightedEntryException):
			Agenda.parse("""
			13:00
			+1:05	foo
			*	blah
			14:00
			""")

	def test_error_weighted_at_beginning(self):
		with self.assertRaises(UnresolvableWeightedEntryException):
			Agenda.parse("""
			*	blah
			14:00
			""")

	def test_error_weighted_at_end(self):
		with self.assertRaises(UnresolvableWeightedEntryException):
			Agenda.parse("""
			14:00
			*	blah
			""")

	def test_error_marker_first(self):
		with self.assertRaises(IllegalAgendaSyntaxException):
			Agenda.parse("""
			:Start
			14:00
			*	blah
			15:00
			""")

	def test_error_relative_out_of_time(self):
		with self.assertRaises(AgendaTimeMismatchException):
			Agenda.parse("""
			13:00
			+1:05	foo
			14:00
			""")

	def test_empty(self):
		agenda = Agenda.parse("")
		self.assertEqual(len(agenda), 0)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>


from pyradium.xmlhooks.XMLHookRegistry import BaseHook, XMLHookRegistry
from pyradium.Tools import XMLTools
from pyradium.Exceptions import NoSuchVariableException, CodeExecutionFailedException

@XMLHookRegistry.register_hook
class VariableSubstitutionHook(BaseHook):
	_TAG_NAME = "sub"

	@classmethod
	def handle(cls, rendered_presentation, node):
		if node.hasAttribute("name"):
			# Direct variable
			varname = node.getAttribute("name")
			if varname not in rendered_presentation.renderer.presentation.variables:
				raise NoSuchVariableException(f"No such variable to substitute in s:sub: {varname}")
			value = rendered_presentation.renderer.presentation.variables[varname]
		else:
			code = XMLTools.inner_text(node)
			try:
				exec_globals = {
					"v":	rendered_presentation.renderer.presentation.variables,
				}
				value = eval(code, exec_globals)
			except Exception as e:
				raise CodeExecutionFailedException(f"Code execution failed in in s:sub \"{code}\" with {e.__class__.__name__}: {str(e)}") from e

		value = str(value)
		replacement_node = node.ownerDocument.createTextNode(value)
		return replacement_node



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import sys
import time
import subprocess
import json
import logging
import traceback
from .BaseAction import BaseAction
from .Presentation import Presentation
from .RenderingParameters import RenderingParameters
from .Renderer import Renderer
from .Exceptions import XMLFileNotFoundException, CallingProcessException, PyRadiumException, DeploymentException
from .Enums import PresentationFeature
from .Deployment import Deployment

_log = logging.getLogger(__spec__.name)

class ActionRender(BaseAction):
	_DEFAULT_PRESENTATION_FEATURES = set([ PresentationFeature.Interactive, PresentationFeature.Timer ])

	def _wait_for_change(self, renderer):
		cmd = [ "inotifywait" ]
		cmd += [ "-q", "-r" ]
		cmd += [ "--exclude", r"\..*\.sw[a-z]" ]
		cmd += [ "-e", "modify", "-e", "close_write", "-e", "create" ]

		sources = [ ]
		sources += renderer.presentation.sources
		sources += renderer.rendering_params.template_dirs
		sources += renderer.rendering_params.include_dirs
		sources += self._args.re_render_watch
		sources = [ source for source in sources if os.path.exists(source) ]
		cmd += sources
		proc = subprocess.run(cmd, check = False)
		if proc.returncode not in [ 0, 1 ]:
			raise CallingProcessException("inotifywait returned with returncode %d." % (proc.returncode))

	def _get_presentation_features(self):
		presentation_features = set(self._DEFAULT_PRESENTATION_FEATURES)
		enabled_features = set(PresentationFeature(x) for x in self._args.enable_presentation_feature)
		disabled_features = set(PresentationFeature(x) for x in self._args.disable_presentation_feature)
		overlap = enabled_features & disabled_features
		if len(overlap) > 0:
			print("Presentation feature can not be enabled and disabled at the same time: %s" % (", ".join(sorted(value.name for value in overlap))), file = sys.stderr)
			return None
		presentation_features |= enabled_features
		presentation_features -= disabled_features

		if (PresentationFeature.Timer in presentation_features) and (PresentationFeature.Interactive not in presentation_features):
			_log.warning("The 'timer' feature implies the 'interactive' feature, which has been selected as well.")
			presentation_features.add(PresentationFeature.Interactive)
		return presentation_features

	def run(self):
		presentation_features = self._get_presentation_features()
		if presentation_features is None:
			return 1

		if len(presentation_features) == 0:
			_log.debug("Rendering without any presentation features.")
		else:
			_log.debug("Rendering with features: %s", ", ".join(sorted(value.name for value in presentation_features)))

		if (not self._args.force) and os.path.exists(self._args.outdir):
			print("Refusing to overwrite: %s" % (self._args.outdir), file = sys.stderr)
			return 1

		if self._args.inject_metadata is None:
			injected_metadata = None
		else:
			with open(self._args.inject_metadata) as f:
				injected_metadata = json.load(f)
				_log.debug("Injected metadata: %s", str(injected_metadata))

		if self._args.resource_dir is not None:
			(resource_dir, resource_uri) = self._args.resource_dir
		else:
			(resource_dir, resource_uri) = (self._args.outdir, "")

		renderer = None
		while True:
			force_wait_secs = None
			try:
				render_success = False
				t0 = time.time()
				rendering_parameters = RenderingParameters(
						template_style = self._args.template_style,
						template_style_opts = self._args.style_option,
						honor_pauses = not self._args.remove_pauses,
						collapse_animation = self._args.collapse_animation,
						extra_template_dirs = self._args.template_dir,
						include_dirs = [ os.path.dirname(self._args.infile) or "." ] + self._args.include_dir,
						index_filename = self._args.index_filename,
						resource_uri = resource_uri,
						geometry = self._args.geometry,
						image_max_dimension = self._args.image_max_dimension,
						presentation_features = presentation_features,
						injected_metadata = injected_metadata,
						trustworthy_source = self._args.trustworthy_source,
						allow_missing_svg_fonts = self._args.allow_missing_svg_fonts)
				presentation = Presentation.load_from_file(self._args.infile, rendering_parameters)
				renderer = Renderer(presentation, rendering_parameters)
				rendered_presentation = renderer.render(resource_directory = resource_dir, deploy_directory = self._args.outdir)
				t1 = time.time()
				_log.info("Successfully rendered presentation into directory \"%s\", took %.1f seconds", self._args.outdir, t1 - t0)

				if self._args.deploy_presentation:
					if "deployment" not in presentation.variables:
						raise DeploymentException("Deployment requested, but no deployment configuration available.")
					deployment = Deployment(deployment_configuration = presentation.variables["deployment"], presentation_directory = self._args.outdir)
					if len(self._args.deployment_name) == 0:
						deployment.deploy("default")
					else:
						for configuration_name in self._args.deployment_name:
							deployment.deploy(configuration_name)
				render_success = True
			except XMLFileNotFoundException as e:
				# This can happen when we save an XML file in VIM and inotify
				# notifies pyradium too quickly (while the file is still not
				# existent). Then we want to re-render quickly to remedy the issue.
				_log.error("Rendering failed: [%s] %s", e.__class__.__name__, str(e))
				force_wait_secs = 1
			except PyRadiumException as e:
				_log.error("Rendering failed: [%s] %s", e.__class__.__name__, str(e))
				if _log.isEnabledFor(logging.DEBUG):
					print(traceback.format_exc())
			if not self._args.re_render_loop:
				break
			if (renderer is None) or (force_wait_secs is not None):
				sleep_duration_secs = force_wait_secs or 5
				_log.warning("Unable to watch files for change since parsing the source was impossible; sleeping for %d seconds instead.", sleep_duration_secs)
				time.sleep(sleep_duration_secs)
			else:
				self._wait_for_change(renderer)
		return 0 if render_success else 1



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import re
import enum
import collections
import logging
from .Exceptions import TimeSpecificationError

_log = logging.getLogger(__spec__.name)

class TimeSpecificationType(enum.IntEnum):
	Relative = 0
	Absolute = 1

class TimeSpecification():
	_ABS_REGEX = re.compile(r"("
							+ r"(?P<single_value>(\d*\.)?\d+)\s*(?P<unit>hrs?|mins?|secs?|h|m|s)"
							+ r"|(?P<hrs>\d+):(?P<mins>\d+):(?P<secs>\d+)"
							+ r"|(?P<val1>\d+):(?P<val2>\d+)\s*(?P<valunit>h:?m|m:?s)?"
							+ r")")
	_REL_REGEX = re.compile(r"(?P<rel_value>(\d*\.)?\d+)")

	def __init__(self, value, spec_type):
		assert(isinstance(spec_type, TimeSpecificationType))
		self._value = value
		self._spec_type = spec_type

	@property
	def relvalue(self):
		assert(self.spec_type == TimeSpecificationType.Relative)
		return self._value

	@property
	def duration_secs(self):
		assert(self.spec_type == TimeSpecificationType.Absolute)
		return self._value

	@property
	def duration_mins(self):
		assert(self.spec_type == TimeSpecificationType.Absolute)
		return self._value / 60

	@property
	def spec_type(self):
		return self._spec_type

	@classmethod
	def parse(cls, abs_string = None, rel_string = None, default_abs_interpretation = None):
		if (abs_string is None) and (rel_string is None):
			raise TimeSpecificationError("Either absolute or relative timing must be supplied.")
		if (abs_string is not None) and (rel_string is not None):
			raise TimeSpecificationError("Either absolute or relative timing must be supplied, not both.")

		if abs_string is not None:
			match = cls._ABS_REGEX.fullmatch(abs_string)
			if match is None:
				raise TimeSpecificationError("Invalid string for absolute timespecification: \"%s\"" % (abs_string))
			match = match.groupdict()
			if match["unit"] is not None:
				# A unit was supplied
				if match["unit"] in [ "hr", "hrs", "h" ]:
					value = float(match["single_value"]) * 3600
				elif match["unit"] in [ "min", "mins", "m" ]:
					value = float(match["single_value"]) * 60
				elif match["unit"] in [ "sec", "secs", "s" ]:
					value = float(match["single_value"])
				else:
					raise NotImplementedError(match["unit"])
			elif match["hrs"] is not None:
				value = (3600 * int(match["hrs"])) + (60 * int(match["mins"])) + int(match["secs"])
			else:
				# xx:yy format, need to infer the correct unit (hrs:mins or
				# mins:secs)
				given_format = match["valunit"] or default_abs_interpretation
				if given_format is None:
					raise TimeSpecificationError("For the given timespec '%s' of format xx:yy it is impossible to tell wether hrs:mins or mins:secs is meant and no default value is available. Please specify." % (abs_string))
				if given_format in [ "h:m", "hm" ]:
					value = (3600 * int(match["val1"])) + (60 * int(match["val2"]))
				elif given_format in [ "m:s", "ms" ]:
					value = (60 * int(match["val1"])) + int(match["val2"])
				else:
					raise NotImplementedError(given_format)
			return cls(spec_type = TimeSpecificationType.Absolute, value = value)
		else:
			match = cls._REL_REGEX.fullmatch(rel_string)
			if match is None:
				raise TimeSpecificationError("Invalid string for relative timespecification: \"%s\"" % (rel_string))
			match = match.groupdict()
			return cls(spec_type = TimeSpecificationType.Relative, value = float(match["rel_value"]))

	def __str__(self):
		if self.spec_type == TimeSpecificationType.Relative:
			return "TimeSpecification<%s, %.1f>" % (self.spec_type.name, self.relvalue)
		elif self.spec_type == TimeSpecificationType.Absolute:
			return "TimeSpecification<%s, %.0f sec>" % (self.spec_type.name, self.duration_secs)
		else:
			return "TimeSpecification<%s>" % (self.spec_type.name)

class PresentationSchedule():
	_TimeSlice = collections.namedtuple("TimeSlice", [ "slide_no", "slide_ratio", "begin_ratio", "end_ratio", "time_seconds" ])

	def __init__(self, presentation_time_seconds = None):
		self._presentation_time_seconds = presentation_time_seconds
		self._time_specs = { }
		self._timeslices = None
		self._max_slide_no = 0

	def have_slide(self, slide_no):
		self._max_slide_no = max(self._max_slide_no, slide_no)

	def set_slide_no(self, slide_no, time_spec):
		if self._timeslices is not None:
			# Already computed, disregards
			return

		assert(isinstance(time_spec, TimeSpecification))
		_log.trace("Setting slide timing of %d to %s", slide_no, str(time_spec))
		self._time_specs[slide_no] = time_spec
		self._max_slide_no = max(self._max_slide_no, slide_no)

	def _slide_range(self):
		return range(1, self._max_slide_no + 1)

	def _slide_no_timespec(self):
		for slide_no in self._slide_range():
			yield (slide_no, self._time_specs.get(slide_no))

	def compute(self):
		if self._timeslices is None:
			abs_sum_secs = 0
			rel_sum_pts = 0
			for (slide_no, time_spec) in self._slide_no_timespec():
				if time_spec is not None:
					if time_spec.spec_type == TimeSpecificationType.Absolute:
						abs_sum_secs += time_spec.duration_secs
					elif time_spec.spec_type == TimeSpecificationType.Relative:
						rel_sum_pts += time_spec.relvalue
					else:
						raise NotImplementedError(time_spec.spec_type)
				else:
					rel_sum_pts += 1

			if (abs_sum_secs > 0) and (self._presentation_time_seconds is None):
				raise TimeSpecificationError("You are free to not define a presentation length beforehand, but then absolute time references may not be used.")

			if self._presentation_time_seconds is not None:
				active_rel_time_secs = self._presentation_time_seconds - abs_sum_secs
				if active_rel_time_secs < 0:
					raise TimeSpecificationError("Presentation active time is %.0f seconds, but %.0f seconds already allocated for absolute slides." % (self._presentation_time_seconds, abs_sum_secs))
			else:
				# Doesn't matter at all, we're only dealing with relative references
				active_rel_time_secs = 1

			if rel_sum_pts == 0:
				# If they are in sum 0, it does not matter anyways, all the
				# numerators will be zero and this way we save two if-elses to
				# avoid divide-by-zero
				rel_sum_pts += 1

			timeslices = { }
			current_slide_ratio = 0
			for (slide_no, time_spec) in self._slide_no_timespec():
				if time_spec is not None:
					if time_spec.spec_type == TimeSpecificationType.Absolute:
						slide_time_secs = time_spec.duration_secs
					elif time_spec.spec_type == TimeSpecificationType.Relative:
						slide_time_secs = time_spec.relvalue / rel_sum_pts * active_rel_time_secs
				else:
					slide_time_secs = 1 / rel_sum_pts * active_rel_time_secs

				if self._presentation_time_seconds is not None:
					slide_ratio = slide_time_secs / self._presentation_time_seconds
				else:
					slide_ratio = slide_time_secs
				timeslice = self._TimeSlice(slide_no = slide_no, slide_ratio = slide_ratio, begin_ratio = current_slide_ratio, end_ratio = current_slide_ratio + slide_ratio, time_seconds = slide_time_secs)
				current_slide_ratio += slide_ratio
				timeslices[slide_no] = timeslice

			self._timeslices = timeslices
		if _log.isEnabledFor(logging.TRACE):
			self.dump()
		return self._timeslices

	def dump(self):
		_log.trace("Dumping %d timeslices determined from presentation schedule:", len(self._timeslices))
		sum_ratios = 0
		sum_duration = 0
		for (slide_no, timeslice) in sorted(self._timeslices.items()):
			sum_ratios += timeslice.slide_ratio
			sum_duration += timeslice.time_seconds
			_log.trace("Slide %2d: ratio %.4f  from %.4f  to %.4f  duration %.0f secs", timeslice.slide_no, timeslice.slide_ratio, timeslice.begin_ratio, timeslice.end_ratio, timeslice.time_seconds)
		_log.trace("Schedule sum of ratios %.3f (should be 1), sum of time %.0f secs", sum_ratios, sum_duration)

	@property
	def slide_ratio_list(self):
		if self._timeslices is None:
			return [ ]
		else:
			return [ timeslice.slide_ratio for (slide_no, timeslice) in sorted(self._timeslices.items()) ]

	def __getitem__(self, slide_no):
		if self._timeslices is None:
			return self._TimeSlice(slide_no = slide_no, slide_ratio = 1, begin_ratio = 0, end_ratio = 1, time_seconds = 1)
		else:
			return self._timeslices[slide_no]

	def __iter__(self):
		return iter(self._timeslices.values())



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import logging
import subprocess

logging.TRACE = logging.DEBUG - 1
logging.SINGLESTEP = logging.DEBUG - 2
logging.addLevelName(logging.TRACE, "TRACE")
logging.addLevelName(logging.SINGLESTEP, "SINGLESTEP")

class CustomLogger(logging.Logger):
	@property
	def subproc_target(self):
		if self.isEnabledFor(logging.TRACE):
			return None
		else:
			return subprocess.DEVNULL

	def trace(self, msg, *args, **kwargs):
		if self.isEnabledFor(logging.TRACE):
			self._log(logging.TRACE, msg, args, **kwargs)

	def singlestep(self, msg, *args, **kwargs):
		if self.isEnabledFor(logging.SINGLESTEP):
			self._log(logging.SINGLESTEP, msg, args, **kwargs)

	@classmethod
	def set_logging_by_verbosity(cls, verbosity):
		if verbosity == 0:
			loglevel = logging.WARN
		elif verbosity == 1:
			loglevel = logging.INFO
		elif verbosity == 2:
			loglevel = logging.DEBUG
		elif verbosity == 3:
			loglevel = logging.TRACE
		else:
			loglevel = logging.SINGLESTEP
		logging.basicConfig(format = "{name:>30s} [{levelname:.1s}]: {message}", style = "{", level = loglevel)

logging.setLoggerClass(CustomLogger)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import tempfile
from pyradium.Tools import ImageTools
from pyradium.DigitalTimingDiagram import DigitalTimingDiagram
from .BaseRenderer import BaseRenderer

@BaseRenderer.register
class DigitalTimingDiagramRenderer(BaseRenderer):
	_NAME = "dtg"

	@property
	def properties(self):
		return {
			"version":			1,
		}

	def render(self, property_dict):
		# Render to SVG first
		constructor_args_names = [ "xdiv", "height", "vertical_distance", "marker_extend", "clock_ticks" ]
		constructor_args = { }
		for arg_name in constructor_args:
			if arg_name in property_dict:
				constructor_args[arg_name] = property_dict[arg_name]

		dtg = DigitalTimingDiagram(**constructor_args)
		dtg.parse_and_write(property_dict["data"])

		with tempfile.NamedTemporaryFile(prefix = "digital_timing_", suffix = ".svg", mode = "w") as svg_file:
			dtg.svg.write(svg_file)
			svg_file.flush()

			# Postprocess in Inkscape to fit the canvas size
			ImageTools.svg_canvas_size_to_object(svg_file.name)

			with open(svg_file.name) as f:
				svg_data = f.read()

		result = {
			"svg":			svg_data,
		}
		return result



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import datetime
import dataclasses
from .Tools import EvalTools
from .Exceptions import InvalidFStringExpressionException, InfiniteRecusionVariableSubstitution, InvalidDateException

class DateTimeWrapper():
	_DEFAULT_PARSE_STRINGS = [
		"%Y-%m-%d",
		"%m/%d/%Y",
		"%d.%m.%Y",
	]

	def __init__(self, dt):
		self._dt = dt

	@classmethod
	def parse(cls, datespec: str, parse_string: str | None = None):
		if parse_string is None:
			parse_strings = cls._DEFAULT_PARSE_STRINGS
		else:
			parse_strings = [ parse_string ]

		for parse_string in parse_strings:
			try:
				return cls(datetime.datetime.strptime(datespec, parse_string))
			except ValueError:
				pass
		raise InvalidDateException(f"Unable to parse datespec: {datespec}")

	def add_days(self, delta_days):
		return DateTimeWrapper(self._dt + datetime.timedelta(delta_days))

	def strftime(self, fmt):
		return self._dt.strftime(fmt)


class VariableSubstitutionContainer():
	@dataclasses.dataclass
	class EvaluatedExpression():
		value: None = None

	def __init__(self, content, environment: dict | None = None, self_varname: str | None = "v"):
		self._content = content
		if environment is None:
			self._environment = { }
		else:
			self._environment = environment
		if self_varname is not None:
			self._environment[self_varname] = self
		self._evaluating_var = set()

	@classmethod
	def default_environment(cls):
		return {
			"int":		int,
			"datetm":	DateTimeWrapper,
		}

	@classmethod
	def merge_dicts(cls, obj1, obj2):
		assert(isinstance(obj1, dict))
		assert(isinstance(obj2, dict))
		result = obj1
		for (key, value) in obj2.items():
			if (key in result) and isinstance(result[key], dict) and isinstance(value, dict):
				result[key] = cls.merge_dicts(result[key], value)
			else:
				result[key] = value
		return result

	def evaluate_fstring(self, fstring):
		if "\"\"\"" in fstring:
			raise InvalidFStringExpressionException(f"Unable to evaluate as f-string: {fstring} may not contain triple quotes")
		expression = "f\"\"\"" + fstring + "\"\"\""
		try:
			return EvalTools.secure_eval(expression, self._environment)
		except Exception as e:
			raise InvalidFStringExpressionException(f"Unable to evaluate as f-string: {expression} -- {type(e).__name__}: {str(e)}") from e

	def evaluate_all(self):
		def _evaluate_if_possible(value):
			if isinstance(value, VariableSubstitutionContainer):
				value = value.evaluate_all()
			return value

		if isinstance(self._content, dict):
			result = { key: _evaluate_if_possible(value) for (key, value) in self.items() }
		elif isinstance(self._content, list):
			result = [ _evaluate_if_possible(item) for item in self ]
		else:
			raise NotImplementedError(type(self._content))
		return result

	def _evaluate(self, key, element):
		if isinstance(element, self.EvaluatedExpression):
			# expression already evaluated, return value
			return element.value
		elif isinstance(element, str):
			# not yet evaluated, evaluate as f-string
			if key in self._evaluating_var:
				raise InfiniteRecusionVariableSubstitution(f"Infinite recursion when accessing dictionary member {key}.")
			self._evaluating_var.add(key)
			evaluated_value = self.evaluate_fstring(fstring = element)
			self._evaluating_var.remove(key)
			self._content[key] = self.EvaluatedExpression(value = evaluated_value)
			return evaluated_value
		elif isinstance(element, (list, dict)):
			return VariableSubstitutionContainer(content = element, environment = self._environment, self_varname = None)
		else:
			return element

	def __contains__(self, element):
		return element in self._content

	def __getitem__(self, key):
		element = self._content[key]
		return self._evaluate(key, element)

	def get(self, key, default_value = None):
		if key not in self._content:
			return default_value
		return self[key]

	def items(self):
		assert(isinstance(self._content, dict))
		for key in self._content:
			yield (key, self[key])

	def __iter__(self):
		if isinstance(self._content, dict):
			yield from self._content
		elif isinstance(self._content, list):
			result = [ ]
			for key in range(len(self._content)):
				yield self[key]
		else:
			raise NotImplementedError(type(self._content))



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from .BaseRenderer import BaseRenderer
from .ExecRenderer import ExecRenderer
from .GraphvizRenderer import GraphvizRenderer
from .ImageRenderer import ImageRenderer
from .LatexFormulaRenderer import LatexFormulaRenderer
from .PlotRenderer import PlotRenderer
from .DigitalTimingDiagramRenderer import DigitalTimingDiagramRenderer
from .QRCodeRenderer import QRCodeRenderer



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2021-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import logging
from pyradium.Controller import BaseController
from pyradium.Exceptions import UsageException, MalformedXMLInputException
from pyradium.svg import SVGTransformation

_log = logging.getLogger(__spec__.name)

def _parse_frame_range(frame_range_str):
	if frame_range_str is None:
		return None
	frames = set()
	try:
		for element in frame_range_str.split(","):
			element = element.split("-")
			if len(element) == 1:
				frames.add(int(element[0]))
			elif len(element) == 2:
				(from_frame, to_frame) = (int(element[0]), int(element[1]))
				frames |= set(range(from_frame, to_frame + 1))
			else:
				raise MalformedXMLInputException("Do not understand how to parse frame range in animation: {element}")
	except ValueError as e:
		raise MalformedXMLInputException("Frame element in animation is not a string: {str(e)}") from e
	return frames

class AnimationController(BaseController):
	_VALID_ANIMATION_MODES = [ "compose", "compose-all", "replace" ]

	def render(self):
		filename = self.slide.get_xml_slide_var("filename")
		if filename is None:
			raise UsageException("Any 'animation' type slide requires a variable called 'filename' to be set.")
		if not filename.lower().endswith(".svg"):
			raise UsageException("Any 'animation' type slide requires an SVG input filename, but found: %s" % (filename))

		animation_mode = self.slide.get_xml_slide_var("mode")
		if animation_mode is None:
			animation_mode = "compose"
		if animation_mode not in self._VALID_ANIMATION_MODES:
			raise UsageException("An 'animation' type slide needs to have a mode set to any of %s, but found: %s" % (", ".join(self._VALID_ANIMATION_MODES), animation_mode))
		frame_range = _parse_frame_range(self.slide.get_xml_slide_var("range"))

		full_filename = self.rendered_presentation.renderer.lookup_include(filename)

		# Determine the number of layers the we should consider first
		svg = SVGTransformation(full_filename)
		if animation_mode == "compose":
			considered_layers = list(svg.visible_layer_ids)
		elif animation_mode in [ "replace", "compose-all" ]:
			considered_layers = list(svg.layer_ids)
		else:
			raise NotImplementedError(animation_mode)

		if len(considered_layers) == 0:
			_log.warning("SVG animation from %s has no layers to be considered.", filename)
			return

		# Then look at all these layers and determine their special handling
		layer_tags = { }
		for layer_id in considered_layers:
			layer = svg.get_layer(layer_id)
			label = layer.label
			if ":" in label:
				(tags, _) = label.split(":", maxsplit = 1)
				layer_tags[layer_id] = set(tags.split(","))

		# Find out which layers are protected
		protected_layer_ids = set()
		for (layer_id, tags) in layer_tags.items():
			if "protect" in tags:
				protected_layer_ids.add(layer_id)

		svg_transforms = [ ]

		# Store commands to hide all layers first
		for layer_id in considered_layers:
			svg_transforms.append({
				"cmd":			"hide_layer",
				"layer_id":		layer_id,
			})

		# Then show them one-by-one and render each
		additional_slide_var_list = [ ]
		previous_layer_id = None
		renderer = self.rendered_presentation.renderer.get_custom_renderer("img")
		for layer_id in considered_layers:
			tags = layer_tags.get(layer_id, set())
			if "reset" in tags:
				# Hide all below layers but those which are protected
				for below_layer_id in considered_layers:
					if below_layer_id == layer_id:
						break
					elif below_layer_id not in protected_layer_ids:
						svg_transforms.append({
							"cmd":			"hide_layer",
							"layer_id":		below_layer_id,
						})

			svg_transforms.append({
				"cmd":			"show_layer",
				"layer_id":		layer_id,
			})
			if (animation_mode == "replace") and (previous_layer_id is not None):
				svg_transforms.append({
					"cmd":			"hide_layer",
					"layer_id":		previous_layer_id,
				})
			if not "nostop" in tags:
				# Actually emit this image
				rendered_image = renderer.render({
					"src":				full_filename,
					"max_dimension":	self.rendered_presentation.renderer.rendering_params.image_max_dimension,
					"svg_transform":	svg_transforms,
				})
				local_filename = "imgs/anim/%s.%s" % (rendered_image.keyhash, rendered_image.data["extension"])
				self.rendered_presentation.add_file(local_filename, rendered_image.data["img_data"])
				additional_slide_var_list.append({
					"image": local_filename,
				})
			previous_layer_id = layer_id

		if frame_range is not None:
			filtered_slide_var_list = [ ]
			for (frame_no, frame) in enumerate(additional_slide_var_list, 1):
				if frame_no in frame_range:
					filtered_slide_var_list.append(frame)
			additional_slide_var_list = filtered_slide_var_list

		if len(additional_slide_var_list) == 0:
			_log.warning("SVG animation from %s rendered into no slides.", filename)
		else:
			_log.debug("SVG animation from %s rendered into %d slides.", filename, len(additional_slide_var_list))
		if not self.rendered_presentation.renderer.rendering_params.collapse_animation:
			yield from self.slide.emit_nocontent_slide(self.rendered_presentation, self.content_containers, additional_slide_var_list)
		else:
			yield from self.slide.emit_nocontent_slide(self.rendered_presentation, self.content_containers, additional_slide_var_list[-1])



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import sys
import argparse
import pyradium
from .MultiCommand import MultiCommand
from .ActionRender import ActionRender
from .ActionShowStyleOpts import ActionShowStyleOpts
from .ActionModify import ActionModify
from .ActionServe import ActionServe
from .ActionAcroAdd import ActionAcroAdd
from .ActionAcroScan import ActionAcroScan
from .ActionAcroSort import ActionAcroSort
from .ActionAcroTex import ActionAcroTex
from .ActionPurge import ActionPurge
from .ActionHashPresentation import ActionHashPresentation
from .ActionDumpMetadata import ActionDumpMetadata
from .ActionSpellcheck import ActionSpellcheck
from .ActionDictAdd import ActionDictAdd
from .ActionTemplateHelper import ActionTemplateHelper
from .ActionStandalone import ActionStandalone
from .modify.BaseModifyCommand import BaseModifyCommand
from .standalone.BaseStandaloneCommand import BaseStandaloneCommand
from .Enums import PresentationFeature
from .GlobalConfig import GlobalConfig

def _geometry(text):
	text = text.split("x", maxsplit = 1)
	if len(text) != 2:
		raise argparse.ArgumentTypeError("Not a valid geometry: %s" % (text))
	return (int(text[0]), int(text[1]))

def _resource_dir(text):
	text = text.split(":", maxsplit = 1)
	if len(text) != 2:
		raise argparse.ArgumentTypeError("Not a valid resource directory/URI combination: %s" % (text))
	return (text[0], text[1])

def main():
	gc = GlobalConfig.read()
	mc = MultiCommand(description = "HTML presentation renderer", trailing_text = "version: pyradium v%s" % (pyradium.VERSION), run_method = True)

	def genparser(parser):
		parser.add_argument("--image-max-dimension", metavar = "pixels", type = int, default = 1920, help = "When rendering imaages, specifies the maximum dimension they're downsized to. The lower this value, the smaller the output files and the lower the quality. Defaults to %(default)d pixels.")
		parser.add_argument("-I", "--include-dir", metavar = "path", action = "append", default = [ ], help = "Specifies an additional include directory in which, for example, images are located which are referenced from the presentation. Can be issued multiple times.")
		parser.add_argument("-R", "--resource-dir", metavar = "path:uripath", type = _resource_dir, help = "Specifies the resource directory both as the actual deployment directory and the URI it has when serving the presentation. By default, the deployment directory of resources is identical to the output directory and the uripath is '.'.")
		parser.add_argument("--template-dir", metavar = "path", action = "append", default = [ ], help = "Specifies an additional template directories in which template style files are located. Can be issued multiple times.")
		parser.add_argument("-t", "--template-style", metavar = "name", default = "antonio", help = "Template style to use. Defaults to %(default)s.")
		parser.add_argument("-o", "--style-option", metavar = "key=value", action = "append", default = [ ], help = "Pass template-style specific options to the renderer. Must always be in the form of \"key=value\". Use the 'showstyleopts' command to find out which keys are supported for a given template.")
		parser.add_argument("-g", "--geometry", metavar = "width x height", type = _geometry, default = "1280x720", help = "Slide geometry, in pixels. Defaults to %(default)s.")
		parser.add_argument("-r", "--remove-pauses", action = "store_true", help = "Ignore all pause directives and just render the final slides.")
		parser.add_argument("--collapse-animation", action = "store_true", help = "Do not render animations as multiple slides, just show one complete slide.")
		parser.add_argument("-i", "--index-filename", metavar = "filename", default = "index.html", help = "Gives the name of the presentation index file. Defaults to %(default)s. Useful if you want to render multiple presentations in one subdirectory.")
		parser.add_argument("-j", "--inject-metadata", metavar = "filename", help = "Gives the option to inject variables into the presentation variable data. Must point to a JSON filename that contains a dictionary. Will take highest precedence. Useful for changing things like the presentation date on the command line.")
		parser.add_argument("-e", "--enable-presentation-feature", choices = [ enumitem.value for enumitem in PresentationFeature ], default = [ ], action = "append", help = "Enable a specific presentation feature. Can be one of %(choices)s and can be given multiple times.")
		parser.add_argument("-d", "--disable-presentation-feature", choices = [ enumitem.value for enumitem in PresentationFeature ], default = [ ], action = "append", help = "Disable a specific presentation feature. Can be one of %(choices)s and can be given multiple times.")
		parser.add_argument("-l", "--re-render-loop", action = "store_true", help = "Stay in a continuous loop, re-rendering the presentation if anything changes.")
		parser.add_argument("-D", "--deploy-presentation", action = "store_true", help = "Deploy presentation after rendering to a remote server using the mechanism indicated in the presentation metadata.")
		parser.add_argument("--deployment-name", metavar = "name", action = "append", default = [ ], help = "Deploy this deployment configuration. Can be specified multiple times to deploy on mutiple targets. Defaults to 'default'.")
		parser.add_argument("--re-render-watch", metavar = "path", action = "append", default = [ ], help = "By default, all include files and the template directory is being watched for changes. This option gives additional files or directories upon change of which the presentation should be re-rendered.")
		parser.add_argument("--trustworthy-source", action = "store_true", help = "By default, the presentation source code is considered not trustworthy and therefore primitives which allow remote code execution (like s:exec) are disabled by default. If you know that the source of your presentation is trustworthy and want to allow it to execute arbitrary code, then specify this parameter.")
		parser.add_argument("--allow-missing-svg-fonts", action = "store_true", help = "If a font is not present on the local system, pyradium aborts instead of rendering an SVG with replaced fonts. This option allows to render the presentation anyways.")
		parser.add_argument("-f", "--force", action = "store_true", help = "Overwrite files in destination directory if they already exist.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity. Can be specified more than once.")
		parser.add_argument("infile", help = "Input XML file of the presentation.")
		parser.add_argument("outdir", help = "Output directory the presentation is put into.")
	mc.register("render", "Render a presentation", genparser, action = ActionRender)

	def genparser(parser):
		parser.add_argument("--template-dir", metavar = "path", action = "append", default = [ ], help = "Specifies an additional template directories in which template style files are located. Can be issued multiple times.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity. Can be specified more than once.")
		parser.add_argument("template_style", help = "Name of the template that should be shown.")
	mc.register("showstyleopts", "Show all options a specific template style supports", genparser, action = ActionShowStyleOpts)

	modify_commands = list(sorted(BaseModifyCommand.get_supported_cmd_list()))
	if len(modify_commands) > 0:
		def genparser(parser):
			parser.add_argument("subcommand", choices = modify_commands, help = "Name of subcommand to call.")
			parser.add_argument("params", nargs = argparse.REMAINDER, help = "Arguments for the respective sub-command.")
		mc.register("modify", "Modify a presentation through one of many sub-commands", genparser, action = ActionModify)

	def genparser(parser):
		parser.add_argument("-b", "--bind-addr", metavar = "addr", type = str, default = "127.0.0.1", help = "Address to bind to. Defaults to %(default)s.")
		parser.add_argument("-p", "--port", metavar = "port", type = int, default = 8123, help = "Port to serve directory under. Defaults to %(default)s.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("dirname", help = "Directory that should be served.")
	mc.register("serve", "Serve a rendered presentation over HTTP", genparser, action = ActionServe)

	def genparser(parser):
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("acrofile", help = "Acronym database JSON file.")
	mc.register("acroadd", "Add an acryonym to the acronym database", genparser, action = ActionAcroAdd, aliases = [ "aadd" ])

	def genparser(parser):
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("acrofile", help = "Acronym database JSON file.")
		parser.add_argument("infile", help = "Input XML file of the presentation.")
	mc.register("acroscan", "Scans a presentation and suggests acronyms that can be added", genparser, action = ActionAcroScan, aliases = [ "ascan" ])

	def genparser(parser):
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("acrofile", help = "Acronym database JSON file.")
	mc.register("acrosort", "Sort an acryonym database", genparser, action = ActionAcroSort, aliases = [ "asort" ])

	def genparser(parser):
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("acrofile", help = "Acronym database JSON file.")
	mc.register("acrotex", "Convert an acronym database to LaTeX format", genparser, action = ActionAcroTex)

	def genparser(parser):
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
	mc.register("purge", "Purge the document cache", genparser, action = ActionPurge)

	def genparser(parser):
		parser.add_argument("-I", "--include-dir", metavar = "path", action = "append", default = [ ], help = "Specifies an additional include directory in which, for example, images are located which are referenced from the presentation. Can be issued multiple times.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("infile", help = "Input XML file of the presentation.")
	mc.register("hash", "Create a hash of a presentation and all dependencies to detect modifications", genparser, action = ActionHashPresentation)

	def genparser(parser):
		parser.add_argument("-p", "--pretty-print", action = "store_true", help = "Pretty print the output JSON data.")
		parser.add_argument("-I", "--include-dir", metavar = "path", action = "append", default = [ ], help = "Specifies an additional include directory in which, for example, images are located which are referenced from the presentation. Can be issued multiple times.")
		parser.add_argument("-j", "--inject-metadata", metavar = "filename", help = "Gives the option to inject variables into the presentation variable data. Must point to a JSON filename and will override the respective metadata fields of the presentation. Useful for changing things like the presentation date on the command line.")
		parser.add_argument("-o", "--outfile", metavar = "filename", default = "-", help = "Write JSON data to this file. May be '-' to write to stdout. Defaults to %(default)s.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("infile", help = "Input XML file of the presentation.")
	mc.register("dumpmeta", "Dump the metadata dictionary in JSON format", genparser, action = ActionDumpMetadata)

	def genparser(parser):
		have_default_spellchecker = gc.has("spellcheck", "uri") or gc.has("spellcheck", "jar")
		group = parser.add_mutually_exclusive_group(required = not have_default_spellchecker)
		group.add_argument("-u", "--uri", metavar = "uri", default = gc.get("spellcheck", "uri"), help = "Connect to this running LanguageTool server URI.")
		group.add_argument("-j", "--jar", metavar = "jarfile", default = gc.get("spellcheck", "jar"), help = "Start a LanguageTool server using this languagetool-server.jar JAR, automatically connect to it and shut it down after use.")
		parser.add_argument("-l", "--language", metavar = "lang", default = "en-US", help = "Spellcheck in this language. Defaults to %(default)s.")
		parser.add_argument("-m", "--mode", choices = [ "print", "vim", "evim", "fulljson" ], default = gc.get("spellcheck", "mode") or "print", help = "Mode in which spellchecking is performed. Can be one of %(choices)s, defaults to %(default)s.")
		parser.add_argument("-o", "--outfile", metavar = "filename", help = "Write output to this file. By default, outputs to stdout.")
		parser.add_argument("--vim", action = "store_true", help = "Run vim to actually spellcheck the file in question.")
		parser.add_argument("-f", "--force", action = "store_true", help = "Overwrite files in destination directory if they already exist.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("infile", help = "Input XML file of the presentation.")
	mc.register("spellcheck", "Spellcheck an XML presentation file", genparser, action = ActionSpellcheck)

	def genparser(parser):
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("infile", help = "Input file, needs to be a quickfix file in evim format.")
	mc.register("dictadd", "Add false-positive spellcheck errors to the dictionary", genparser, action = ActionDictAdd)

	def genparser(parser):
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("template_name", help = "Name of the template to insert.")
	mc.register("template-helper", "Show different templates on stdout; used in conjunction with vim plugins", genparser, action = ActionTemplateHelper)

	standalone_commands = list(sorted(BaseStandaloneCommand.get_supported_cmd_list()))
	if len(standalone_commands) > 0:
		def genparser(parser):
			parser.add_argument("subcommand", choices = standalone_commands, help = "Name of subcommand to call.")
			parser.add_argument("params", nargs = argparse.REMAINDER, help = "Arguments for the respective sub-command.")
		mc.register("standalone", "Run one of various standalong tools through sub-commands", genparser, action = ActionStandalone)

	return mc.run(sys.argv[1:])

if __name__ == "__main__":
	sys.exit(main())



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyradium.xmlhooks.XMLHookRegistry import BaseHook, XMLHookRegistry
from pyradium.Tools import XMLTools
from pyradium.BooleanExpression import BooleanExpressionParser, TexBooleanExpressionPrinter

@XMLHookRegistry.register_hook
class BoolHook(BaseHook):
	_TAG_NAME = "bool"

	@classmethod
	def handle(cls, rendered_presentation, node):
		properties = {
			"invert_overline":		XMLTools.get_bool_attr(node, "invert_overline", default_value = True),
		}

		expression = XMLTools.inner_text(node)
		tree = BooleanExpressionParser(expression).parse()

		printer = TexBooleanExpressionPrinter(invert_by_overline = properties["invert_overline"])
		tex_equation = printer.print(tree)

		replacement_node = node.ownerDocument.createElement("s:tex")
		replacement_node.appendChild(node.ownerDocument.createTextNode(tex_equation))
		for copy_attr in [ "long", "indent" ]:
			if node.hasAttribute(copy_attr):
				replacement_node.setAttribute(copy_attr, node.getAttribute(copy_attr))

		return replacement_node



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import xml.dom.minidom
from pyradium.Tools import XMLTools
from pyradium.Exceptions import InvalidTransformationException
from .SVGLayer import SVGLayer

class SVGTransformation():
	def __init__(self, svg_filename):
		self._xml = xml.dom.minidom.parse(svg_filename)
		self._svg = XMLTools.findfirst(self._xml, "svg")
		self._layers = [ ]
		self._layers_by_id = { }
		for group_node in XMLTools.findall(self._svg, "g", namespace_uri = "http://www.w3.org/2000/svg"):
			groupmode = group_node.getAttributeNS("http://www.inkscape.org/namespaces/inkscape", "groupmode")
			if groupmode == "layer":
				# Inkscape layer node
				layer = SVGLayer(group_node)
				self._layers.append(layer)
				self._layers_by_id[layer.layer_id] = layer

	@property
	def visible_layer_ids(self):
		for layer in self._layers:
			if layer.is_visible:
				yield layer.layer_id

	@property
	def layer_ids(self):
		for layer in self._layers:
			yield layer.layer_id

	@property
	def layers(self):
		return iter(self._layers)

	def get_layer(self, layer_id):
		return self._layers_by_id[layer_id]

	def _format_text(self, variables):
		for text_node in XMLTools.findall_recurse(self._xml, "text"):
			for tspan_node in XMLTools.findall(text_node, "tspan"):
				for cdata_node in XMLTools.findall_text(tspan_node, recursive = True):
					try:
						replacement_text = cdata_node.wholeText.format(**variables)
					except KeyError as e:
						raise InvalidTransformationException(f"Requested variable substitution is missing variable name: {e}") from e
					cdata_node.replaceWholeText(replacement_text)

	def apply(self, transformation_dict):
		if transformation_dict["cmd"] == "show_layer":
			self.get_layer(transformation_dict["layer_id"]).show()
		elif transformation_dict["cmd"] == "hide_layer":
			self.get_layer(transformation_dict["layer_id"]).hide()
		elif transformation_dict["cmd"] == "format_text":
			self._format_text(transformation_dict["variables"])
		else:
			raise NotImplementedError(transformation_dict["cmd"])

	def apply_all(self, transformation_dicts):
		for transformation_dict in transformation_dicts:
			self.apply(transformation_dict)

	def write(self, f):
		self._xml.writexml(f)

	def write_file(self, filename):
		with open(filename, "w") as f:
			self.write(f)

if __name__ == "__main__":
	svg = SVGTransformation("animation_l2hid.svg")
	print(list(svg.layer_ids))
	layer_ids = list(svg.visible_layer_ids)
	print(layer_ids)
	for layer_id in layer_ids:
		svg.get_layer(layer_id).hide()
	svg.get_layer(layer_ids[0]).show()
	svg.apply({
		"cmd": "replace_text",
		"search": "MUH TEXT",
		"replace": "This was Muh.",
	})
	svg.apply({
		"cmd": "replace_text",
		"search": "FOO TEXT",
		"replace": "This was Foo.",
	})
	svg.write_file("/tmp/x.svg")



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

class SVGStyle():
	def __init__(self, style_dict: dict | None = None, auto_sync: bool = False, node = None):
		assert((style_dict is None) or isinstance(style_dict, dict))
		self._style = style_dict if (style_dict is not None) else { }
		self._auto_sync = auto_sync
		self._node = node

	@property
	def auto_sync(self):
		return self._auto_sync

	@auto_sync.setter
	def auto_sync(self, value):
		self._auto_sync = value

	@property
	def node(self):
		return self._node

	@node.setter
	def node(self, value):
		self._node = value

	@classmethod
	def from_style_str(cls, style_str: str):
		style_dict = { }
		for style_item in style_str.split(";"):
			style_item = style_item.strip()
			if style_item == "":
				continue
			style_item = style_item.split(":", maxsplit = 1)
			if len(style_item) == 2:
				(key, value) = style_item
				key = key.strip()
				value = value.strip()
				style_dict[key] = value
		return cls(style_dict)

	@classmethod
	def from_node(cls, node, auto_sync: bool = False):
		if node.hasAttribute("style"):
			style = cls.from_style_str(node.getAttribute("style"))
		else:
			style = cls()
		style.node = node
		style.auto_sync = auto_sync
		return style

	def sync_to_node(self):
		if (len(self._style) == 0) and self._node.hasAttribute("style"):
			self._node.removeAttribute("style")
		else:
			self._node.setAttribute("style", self.serialize())

	@property
	def is_visible(self):
		if "display" in self._style:
			hidden = (self["display"] == "none")
			return not hidden
		return True

	def serialize(self):
		return ";".join("%s:%s" % (key, value) for (key, value) in self._style.items())

	def _on_style_changed(self):
		if (self._node is not None) and self._auto_sync:
			self.sync_to_node()

	def update(self, style_dict: dict):
		for (key, value) in style_dict.items():
			self._style[key] = value
		self._on_style_changed()

	def default_path(self):
		self.update({
			"fill":				"none",
			"stroke":			"#000000",
			"stroke-width":		"1px",
			"stroke-linecap":	"butt",
			"stroke-linejoin":	"miter",
			"stroke-opacity":	"1",
		})
		return self

	def default_text(self):
		self.update({
			"font-style":		"normal",
			"font-weight":		"normal",
			"font-size":		"12px",
			"line-height":		"1.25",
			"font-family":		"sans-serif",
			"white-space":		"pre",
			"fill":				"#000000",
			"fill-opacity":		"1",
			"stroke":			"none",
		})
		return self

	def __setitem__(self, key: str, value: str):
		self._style[key] = value
		self._on_style_changed()

	def __getitem__(self, key: str):
		return self._style.get(key)

	def __str__(self):
		return f"SVGStyle<{self._style}>"



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
import logging
import functools
import subprocess
import xml.dom.minidom
from pyradium.Tools import XMLTools
from pyradium.Exceptions import SVGValidationError
from .SVGStyle import SVGStyle

_log = logging.getLogger(__spec__.name)

class SVGValidatorErrorClass(enum.IntEnum):
	CheckDisabled = 0
	EmitWarning = 1
	ThrowException = 2

class SVGValidator():
	def __init__(self, check_missing_fonts: SVGValidatorErrorClass = SVGValidatorErrorClass.EmitWarning):
		self._check_missing_fonts = check_missing_fonts

	@functools.lru_cache(maxsize = 1000)
	def _have_font(self, font_family):
		return len(subprocess.check_output([ "fc-list", font_family ])) > 0

	def _validate_fonts(self, svg_filename, root):
		found_missing = set()
		for tagname in [ "text", "tspan" ]:
			for node in XMLTools.findall_recurse(root, tagname):
				style = SVGStyle.from_node(node)
				font_family = style["font-family"]
				if font_family is None:
					continue
				if (font_family[0] == "'") and (font_family[-1] == "'") and (len(font_family) >= 2):
					font_family = font_family[1 : -1]

				if font_family in found_missing:
					# Already reported.
					continue

				if not self._have_font(font_family):
					found_missing.add(font_family)
					if self._check_missing_fonts == SVGValidatorErrorClass.EmitWarning:
						_log.warning("SVG file %s has missing font: %s", svg_filename, font_family)
					elif self._check_missing_fonts == SVGValidatorErrorClass.ThrowException:
						raise SVGValidationError(f"SVG file {svg_filename} has missing font: {font_family}")

	def validate(self, svg_filename):
		root = xml.dom.minidom.parse(svg_filename)
		if self._check_missing_fonts != SVGValidatorErrorClass.CheckDisabled:
			self._validate_fonts(svg_filename, root)

if __name__ == "__main__":
	validator = SVGValidator()
	validator.validate("examples/3dbox.svg")



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyradium.xmlhooks.XMLHookRegistry import BaseHook, XMLHookRegistry
from pyradium.Tools import XMLTools
from pyradium.svg import SVGStyle

@XMLHookRegistry.register_hook
class TexHook(BaseHook):
	_TAG_NAME = "tex"

	@classmethod
	def handle(cls, rendered_presentation, node):
		properties = {
			"formula":	XMLTools.inner_text(node),
			"long":		XMLTools.get_bool_attr(node, "long"),
		}
		if node.hasAttribute("scale"):
			user_scale = float(node.getAttribute("scale"))
		else:
			user_scale = 1
		tex_renderer = rendered_presentation.renderer.get_custom_renderer("latex")
		rendered_formula = tex_renderer.render(properties)
		local_filename = "imgs/latex/%s.png" % (rendered_formula.keyhash)
		uri = "%simgs/latex/%s.png" % (rendered_presentation.renderer.rendering_params.resource_uri, rendered_formula.keyhash)

		scale_factor = 0.625 * user_scale
		width_px = round(rendered_formula.data["info"]["width"] * scale_factor)
		baseline_px = round(rendered_formula.data["info"]["baseline"] * scale_factor)
		#print(properties["formula"], rendered_formula.data["info"], width_px)

		# For now, always wrap in div for long formulae. Could easily add an
		# option to override this if it's useful.
		wrap_in_div = properties["long"]

		img_node = node.ownerDocument.createElement("img")
		img_node.setAttribute("src", uri)
		img_node.setAttribute("alt", properties["formula"])

		img_style = SVGStyle.from_node(img_node)
		img_style["width"] = f"{width_px}px"
		img_style["margin-top"] = "5px"
		if not properties["long"]:
			img_style["margin-bottom"] = f"{-baseline_px + 1}px"
		if node.hasAttribute("indent"):
			indent = float(node.getAttribute("indent"))
			indent_px = round(indent * 50)
			img_style["margin-left"] = f"{indent_px}px"
		img_style.sync_to_node()

		if not wrap_in_div:
			replacement_node = img_node
		else:
			replacement_node = node.ownerDocument.createElement("div")
			replacement_node.setAttribute("class", "texformula")
			replacement_node.appendChild(img_node)

		rendered_presentation.add_file(local_filename, rendered_formula.data["png_data"])
		return replacement_node



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2021-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import logging
import subprocess
from pyradium.Tools import HashTools
from .BaseRenderer import BaseRenderer

_log = logging.getLogger(__spec__.name)

@BaseRenderer.register
class GraphvizRenderer(BaseRenderer):
	_NAME = "graphviz"

	@property
	def properties(self):
		return {
			"version":			1,
		}

	def _render_plot_png(self, src, scale):
		with open(src) as f:
			source = f.read()
		dpi = round(250 * scale)
		png_data = subprocess.check_output([ "dot", "-Tpng", f"-Gdpi={dpi}" ], input = source.encode("utf-8"), stderr = _log.subproc_target)
		return png_data

	def rendering_key(self, property_dict):
		src = property_dict["src"]
		return {
			"srchash":		HashTools.hash_file(src),
		}

	def render(self, property_dict):
		src = property_dict["src"]
		scale = float(property_dict.get("scale", "1"))
		img_data = self._render_plot_png(src, scale)
		image = {
			"extension":	"png",
			"img_data":		img_data,
		}
		return image

if __name__ == "__main__":
	from pyradium.RendererCache import RendererCache
	renderer = RendererCache(GraphvizRenderer())
	output = renderer.render({
		"src":	"/tmp/x.dot"
	})
	print(output)
	with open("/tmp/x.png", "wb") as f:
		f.write(output.data["img_data"])



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
import collections
import dataclasses
from pyradium.svg import SVGWriter

class DigitalTimingType(enum.Enum):
	Low = "0"
	High = "1"
	LowHigh = ":"
	LowHighTransition = "!"
	HighZ = "Z"
	Marker = "|"
	Empty = " "

@dataclasses.dataclass
class DigitalTimingCmd():
	cmdtype: DigitalTimingType
	argument: "typing.Any" = None

	@classmethod
	def parse_sequence(cls, text):
		sequence = [ ]
		index = 0
		while index < len(text):
			char = text[index]
			index += 1
			match char:
				case "0":
					cmd = DigitalTimingCmd(cmdtype = DigitalTimingType.Low)
				case "1":
					cmd = DigitalTimingCmd(cmdtype = DigitalTimingType.High)
				case ":":
					cmd = DigitalTimingCmd(cmdtype = DigitalTimingType.LowHigh)
				case "!":
					cmd = DigitalTimingCmd(cmdtype = DigitalTimingType.LowHighTransition)
				case "Z":
					cmd = DigitalTimingCmd(cmdtype = DigitalTimingType.HighZ)
				case "|":
					marker_label = None
					if (index < len(text)) and (text[index] == "'"):
						# Label is present
						marker_label = ""
						index += 1
						while index < len(text):
							if text[index] == "'":
								break
							else:
								marker_label += text[index]
							index += 1
						index += 1
					cmd = DigitalTimingCmd(cmdtype = DigitalTimingType.Marker, argument = marker_label)
				case "_":
					cmd = DigitalTimingCmd(cmdtype = DigitalTimingType.Empty)
				case " ":
					continue
				case _:
					raise NotImplementedError(f"Unknown character in sequence diagram: {char}")
			sequence.append(cmd)
		return sequence

class DigitalTimingDiagram():
	_Marker = collections.namedtuple("Marker", [ "x", "label" ])

	def __init__(self, xdiv = 10, height = 30, vertical_distance = 10, marker_extend = 20, clock_ticks = True, low_high_lines = False):
		self._xdiv = xdiv
		self._height = height
		self._vertical_distance = vertical_distance
		self._marker_extend = marker_extend
		self._render_clock_ticks = clock_ticks
		self._low_high_lines = low_high_lines
		self._risefall = height / 8
		self._svg = SVGWriter()
		self._path = None
		self._plot_count = 0
		self._clock_ticks = 0
		if self._low_high_lines:
			self._svg.group("low_high_horizontal")
		if self._render_clock_ticks:
			self._svg.group("clock_ticks")
		self._markers = [ ]

	@property
	def svg(self):
		return self._svg

	@property
	def base_height(self):
		return ((self._height + self._vertical_distance) * self._plot_count) - self._vertical_distance

	def _transition_middle(self, y, transition_scale = 1):
		transition_width = transition_scale * self._risefall * (abs(y) / self._height)
		lead = (self._xdiv - transition_width) / 2
		self._path.horiz_rel(lead)
		self._path.line_rel(transition_width, y)
		self._path.horiz_rel(lead)

	def _render_signal_sequence(self, signal_name, x, y, cmds):
		prev = None
		self._plot_count += 1
		abs_y_mid = y + (self._height / 2)
		self._path = self._svg.new_path(x, abs_y_mid, group_name = "signal")

		text_width = 50
		svg_text = self._svg.new_text_span(x - text_width, abs_y_mid - 6, text_width, 30, signal_name.lstrip("!"), group_name = "signal")
		svg_text.style["text-align"] = "right"
		svg_text.style["font-family"] = "'Latin Modern Roman'"
		if signal_name.startswith("!"):
			# text-decoration: overline does not work reliably, emulate
			svg_text = self._svg.new_text_span(x - text_width, abs_y_mid - 6 - 12, text_width, 30, "_" * (len(signal_name) - 1), group_name = "signal")
			svg_text.style["text-align"] = "right"
			svg_text.style["font-family"] = "'Latin Modern Roman'"

		for cur in cmds:
			if prev is None:
				prev = cur
				if prev.cmdtype in [ DigitalTimingType.Low, DigitalTimingType.LowHigh, DigitalTimingType.LowHighTransition ]:
					self._path.move_rel(0, self._height / 2)
				elif prev.cmdtype in [ DigitalTimingType.High ]:
					self._path.move_rel(0, -self._height / 2)

			match (prev.cmdtype, cur.cmdtype):
				case (DigitalTimingType.Low, DigitalTimingType.Low) | (DigitalTimingType.High, DigitalTimingType.High) | (DigitalTimingType.HighZ, DigitalTimingType.HighZ):
					self._path.horiz_rel(self._xdiv)

				case (DigitalTimingType.Low, DigitalTimingType.High):
					self._transition_middle(-self._height)

				case (DigitalTimingType.High, DigitalTimingType.Low):
					self._transition_middle(self._height)

				case (DigitalTimingType.Low, DigitalTimingType.HighZ) | (DigitalTimingType.HighZ, DigitalTimingType.High):
					self._transition_middle(-self._height / 2)

				case (DigitalTimingType.High, DigitalTimingType.HighZ) | (DigitalTimingType.HighZ, DigitalTimingType.Low):
					self._transition_middle(self._height / 2)

				case (DigitalTimingType.LowHigh, DigitalTimingType.LowHigh) | (DigitalTimingType.LowHighTransition, DigitalTimingType.LowHigh):
					with self._path.returnto():
						# High to high
						self._path.move_rel(0, -self._height)
						self._path.horiz_rel(self._xdiv)
					# Low to low
					self._path.horiz_rel(self._xdiv)

				case (DigitalTimingType.High, DigitalTimingType.LowHigh):
					with self._path.returnto():
						# High to high
						self._path.horiz_rel(self._xdiv)
					# High to low
					self._transition_middle(self._height)

				case (DigitalTimingType.Low, DigitalTimingType.LowHigh):
					with self._path.returnto():
						# Low to high
						self._transition_middle(-self._height)
					# Low to low
					self._path.horiz_rel(self._xdiv)


				case (DigitalTimingType.LowHigh, DigitalTimingType.High):
					with self._path.returnto():
						# High to high
						self._path.move_rel(0, -self._height)
						self._path.horiz_rel(self._xdiv)
					# Low to low
					self._transition_middle(-self._height)

				case (DigitalTimingType.LowHigh, DigitalTimingType.Low):
					with self._path.returnto():
						# High to low
						self._path.move_rel(0, -self._height)
						self._transition_middle(self._height)
					# Low to low
					self._path.horiz_rel(self._xdiv)

				case (DigitalTimingType.HighZ, DigitalTimingType.LowHigh):
					with self._path.returnto():
						# HighZ to High
						self._transition_middle(-self._height / 2)
					# HighZ to Low
					self._transition_middle(self._height / 2)

				case (DigitalTimingType.LowHigh, DigitalTimingType.HighZ):
					with self._path.returnto():
						# Low to HighZ
						self._transition_middle(-self._height / 2)
					# High to HighZ
					self._path.move_rel(0, -self._height)
					self._transition_middle(self._height / 2)

				case (DigitalTimingType.LowHigh, DigitalTimingType.LowHighTransition) | (DigitalTimingType.LowHighTransition, DigitalTimingType.LowHighTransition):
					with self._path.returnto():
						self._transition_middle(-self._height, transition_scale = 2)
					self._path.move_rel(0, -self._height)
					self._transition_middle(self._height, transition_scale = 2)

				case (_, DigitalTimingType.Empty):
					self._path.move_to(self._path.pos.x, abs_y_mid)
					prev = None
					continue

				case (_, DigitalTimingType.Marker):
					mid_x = self._path.pos.x
					self._markers.append(self._Marker(x = mid_x + self._xdiv / 2, label = cur.argument))
					continue

				case _ as transition:
					raise NotImplementedError(f"Unsupported digital sequence diagram transition: {transition}")
			prev = cur
		self._clock_ticks = max(self._clock_ticks, round(self._path.pos.x / self._xdiv))

	def _render_markers(self):
		for marker in self._markers:
			have_label = (marker.label is not None) and (marker.label != "")
			marker_length = self.base_height if (not have_label) else (self.base_height + self._marker_extend)

			path = self._svg.new_path(marker.x, 0, group_name = "marker")
			path.vert_rel(marker_length)
			path.style["stroke-width"] = 0.5

			if have_label:
				text_width = 100
				text_height = 50

				svg_text = self._svg.new_text_span(marker.x - (text_width / 2), marker_length, text_width, text_height, marker.label, group_name = "marker")
				svg_text.style["text-align"] = "center"

	def _do_render_clock_ticks(self):
		if not self._render_clock_ticks:
			return

		for tick in range(self._clock_ticks):
			x = (tick * self._xdiv) + self._xdiv / 2
			path = self._svg.new_path(x, 0, group_name = "clock_ticks")
			path.vert_rel(self.base_height)
			path.style["stroke-width"] = 0.25
			path.style["stroke"] = "#95a5a6"
			path.style["stroke-miterlimit"] = 4
			path.style["stroke-dasharray"] = "0.75,0.25"
			path.style["stroke-dashoffset"] = 0

	def _do_render_low_high_lines(self):
		x_width = self._clock_ticks * self._xdiv
		for plot in range(self._plot_count):
			y_high = (self._height + self._vertical_distance) * plot
			y_low = y_high + self._height
			for y in [ y_low, y_high ]:
				path = self._svg.new_path(0, y, group_name = "low_high_horizontal")
				path.horiz_rel(x_width)

				path.style["stroke-width"] = 0.5
				path.style["stroke"] = "#95a5a6"
				path.style["stroke-miterlimit"] = 4
				path.style["stroke-dasharray"] = "0.75,0.25"
				path.style["stroke-dashoffset"] = 0

	def parse_and_write(self, text):
		text = text.strip("\r\n")
		varno = 0
		for line in text.split("\n"):
			line = line.strip("\r\n \t")
			if line == "":
				continue
			if line.startswith("#"):
				continue
			(varname, sequence) = line.split("=", maxsplit = 1)
			varname = varname.strip("\t ")
			sequence = sequence.strip("\t ")
			cmds = DigitalTimingCmd.parse_sequence(sequence)
			self._render_signal_sequence(varname, 0, (self._height + self._vertical_distance) * varno, cmds)
			varno += 1
		self._render_markers()
		self._do_render_clock_ticks()
		self._do_render_low_high_lines()



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import re
import collections
from pyradium.Exceptions import IllegalAgendaSyntaxException, UndefinedAgendaTimeException, UnresolvableWeightedEntryException, AgendaTimeMismatchException

AgendaItem = collections.namedtuple("AgendaItem", [ "start_time", "end_time", "duration", "text" ])
_UnresolvedAgendaItem = collections.namedtuple("UnresolvedAgendaItem", [ "spec_type", "value", "text", "markers" ])

class Agenda():
	_AGENDA_REGEX = re.compile(r"((?P<relative>\+)?((?P<day>\d+)\+)?(?P<hour>\d{1,2}):(?P<minute>\d{1,2})(/(?P<divider>\d+(\.\d+)?))?|(?P<wildcard>\*)(?P<weight>\d+(\.\d+)?)?)(\s+(?P<text>.*))?|:(?P<marker>.+)")

	def __init__(self, agenda_items: list[AgendaItem], markers: collections.OrderedDict, name: str | None = None):
		self._agenda_items = agenda_items
		self._markers = markers
		self._name = name

	@property
	def name(self):
		return self._name

	@property
	def markers(self):
		return iter(self._markers.items())

	def dump(self):
		print(f"Agenda with {len(self)} items:")
		for item in self:
			print(f"    {item}")

	def __len__(self):
		return len(self._agenda_items)

	def __getitem__(self, index):
		return self._agenda_items[index]

	@classmethod
	def _hmstr(cls, minutes):
		minutes = round(minutes) % (24 * 60)
		return f"{minutes // 60}:{minutes % 60:02d}"

	@classmethod
	def _roundto(cls, value, boundary):
		return round(value / boundary) * boundary

	@classmethod
	def _resolve_relative_references(cls, unresolved_agenda, reverse = False):
		if len(unresolved_agenda) == 0:
			return [ ]

		now = None
		resolved_items = [ ]
		unresolved_agenda = unresolved_agenda if (not reverse) else reversed(unresolved_agenda)

		for item in unresolved_agenda:
			if item.spec_type == "abs":
				now = item.value
				resolved_items.append(item)
			elif (item.spec_type == "rel") and (now is not None):
				if not reverse:
					now += item.value
				resolved_item = _UnresolvedAgendaItem(spec_type = "abs", value = now, text = item.text, markers = item.markers)
				if reverse:
					now -= item.value
				resolved_items.append(resolved_item)
			else:
				now = None
				resolved_items.append(item)

		if reverse and (now is not None):
			resolved_item = _UnresolvedAgendaItem(spec_type = "abs", value = now, text = None, markers = [ ])
			resolved_items.append(resolved_item)

		if reverse:
			resolved_items.reverse()
		return resolved_items

	@classmethod
	def _resolve_forward_relative_references(cls, unresolved_agenda: list[_UnresolvedAgendaItem]):
		return cls._resolve_relative_references(unresolved_agenda)

	@classmethod
	def _resolve_backward_relative_references(cls, unresolved_agenda: list[_UnresolvedAgendaItem]):
		return cls._resolve_relative_references(unresolved_agenda, reverse = True)

	@classmethod
	def _resolve_all_relative_references(cls, unresolved_agenda: list[_UnresolvedAgendaItem]):
		unresolved_agenda = cls._resolve_forward_relative_references(unresolved_agenda)
		unresolved_agenda = cls._resolve_backward_relative_references(unresolved_agenda)
		return unresolved_agenda

	@classmethod
	def _resolve_enclosed_weighted_references(cls, enclosed_agenda: list[_UnresolvedAgendaItem]):
		assert(len(enclosed_agenda) >= 3)
		assert(enclosed_agenda[0].spec_type == "abs")
		assert(enclosed_agenda[-1].spec_type == "abs")
		resolved_items = [ ]
		total_time = enclosed_agenda[-1].value - enclosed_agenda[0].value
		relative_time = 0
		weight_sum = 0
		for item in enclosed_agenda:
			if item.spec_type == "rel":
				relative_time += item.value
			elif item.spec_type == "weight":
				weight_sum += item.value
		if weight_sum == 0:
			raise UnresolvableWeightedEntryException("Sum of weights is zero.")
		if relative_time > total_time:
			raise UnresolvableWeightedEntryException(f"Unresolvable weighted exception: relative time within enclosed absolute time bracket too long relative {cls._hmstr(relative_time)} but total time only {cls._hmstr(total_time)}")

		weighted_time = total_time - relative_time
		now = enclosed_agenda[0].value
		for item in enclosed_agenda[1 : -1]:
			if item.spec_type == "abs":
				pass
			elif item.spec_type == "rel":
				now += item.value
				item = _UnresolvedAgendaItem(spec_type = "abs", value = now, text = item.text, markers = item.markers)
			elif item.spec_type == "weight":
				now += (item.value / weight_sum) * weighted_time
				item = _UnresolvedAgendaItem(spec_type = "abs", value = now, text = item.text, markers = item.markers)
			resolved_items.append(item)
		return resolved_items

	@classmethod
	def _resolve_weighted_references(cls, unresolved_agenda: list[_UnresolvedAgendaItem]):
		resolved_items = [ ]

		index = 0
		while index < len(unresolved_agenda):
			item = unresolved_agenda[index]
			if item.spec_type == "weight":
				if index == 0:
					raise UnresolvableWeightedEntryException("Unresolvable weighted exception: May not have a weighted relative item at the beginning")
				enclosed = [ unresolved_agenda[index - 1], item ]

				index += 1
				while index < len(unresolved_agenda):
					item = unresolved_agenda[index]
					enclosed.append(item)
					if item.spec_type == "abs":
						break
					index += 1

				if enclosed[-1].spec_type != "abs":
					raise UnresolvableWeightedEntryException("Unresolvable weighted exception: May not have a weighted relative item at the end")
				resolved_items += cls._resolve_enclosed_weighted_references(enclosed)
			else:
				resolved_items.append(item)
				index += 1
		return resolved_items

	@classmethod
	def _resolve(cls, unresolved_agenda: list[_UnresolvedAgendaItem], granularity_minutes: int):
		unresolved_agenda = cls._resolve_all_relative_references(unresolved_agenda)
		unresolved_agenda = cls._resolve_weighted_references(unresolved_agenda)

		resolved_agenda_items = [ ]
		markers = collections.OrderedDict()
		prev_item = None
		for item in unresolved_agenda:
			if (prev_item is not None) and (item.value is not None) and (prev_item.value is not None) and (item.value < prev_item.value):
				raise AgendaTimeMismatchException(f"Time cannot move backwards: Previously at {cls._hmstr(prev_item.value)}, but requested moving back to {cls._hmstr(item.value)} (from {prev_item} to {item})")
			if item.text is not None:
				if (prev_item is None) or (prev_item.value is None):
					raise UndefinedAgendaTimeException(f"Unable to determine start time of event: {item}")

				start = cls._roundto(prev_item.value, granularity_minutes)
				end = cls._roundto(item.value, granularity_minutes)
				duration = end - start
				agenda_item = AgendaItem(start_time = cls._hmstr(start), end_time = cls._hmstr(end), duration = cls._hmstr(duration), text = item.text)
				resolved_agenda_items.append(agenda_item)
			for marker in item.markers:
				ts = cls._roundto(item.value, granularity_minutes)
				markers[marker] = cls._hmstr(ts)
			prev_item = item
		return (resolved_agenda_items, markers)

	@classmethod
	def parse(cls, text: str, name: str | None = None, granularity_minutes: int = 5):
		now = None

		unresolved_items = [ ]
		day = 0
		for line in text.split("\n"):
			line = line.strip()
			if line == "":
				continue

			rematch = cls._AGENDA_REGEX.fullmatch(line)
			if rematch is None:
				raise IllegalAgendaSyntaxException(f"Do not understand agenda element: {line}")

			rematch = rematch.groupdict()
			if rematch["hour"] is not None:
				if rematch["day"] is not None:
					day = int(rematch["day"])
				time_value = (int(rematch["hour"]) * 60) + int(rematch["minute"]) + (1440 * day)
				relative = rematch["relative"] is not None
				if rematch["divider"] is not None:
					if not relative:
						raise IllegalAgendaSyntaxException(f"Specifying a divider for an absolute time element does not make sense: {line}")
					divider = float(rematch["divider"])
					time_value /= divider
				timespec = ("rel" if relative else "abs", time_value)
			elif rematch["wildcard"] is not None:
				if rematch["weight"] is not None:
					weight = float(rematch["weight"])
				else:
					weight = 1
				timespec = ("weight", weight)
			elif rematch["marker"] is not None:
				if len(unresolved_items) == 0:
					raise IllegalAgendaSyntaxException(f"Cannot have a marker as the first element: {line}")
				last_item = unresolved_items[-1]
				last_item.markers.append(rematch["marker"])
				continue

			unresolved_item = _UnresolvedAgendaItem(spec_type = timespec[0], value = timespec[1], text = rematch["text"], markers = [ ])
			unresolved_items.append(unresolved_item)
		(agenda_items, markers) = cls._resolve(unresolved_items, granularity_minutes = granularity_minutes)
		return cls(agenda_items = agenda_items, markers = markers, name = name)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import json
import logging
from .BaseAction import BaseAction
from .RenderingParameters import RenderingParameters
from .StyleParameters import StyleParameters

_log = logging.getLogger(__spec__.name)

class ActionShowStyleOpts(BaseAction):
	def run(self):
		rendering_parameters = RenderingParameters(
			template_style = self._args.template_style,
			extra_template_dirs = self._args.template_dir,
		)
		config_filename = rendering_parameters.template_dirs.lookup(rendering_parameters.template_style + "/configuration.json")
		_log.debug("Using configuration: %s", config_filename)
		with open(config_filename) as f:
			config = json.load(f)

		if "parameters" not in config:
			print("Template has no parameters.")
		else:
			parameters = StyleParameters(config["parameters"])
			parameters.print()



class Vector2D():
	def __init__(self, x = 0, y = 0):
		self._x = x
		self._y = y

	@property
	def x(self):
		return self._x

	@property
	def y(self):
		return self._y

	def __iadd__(self, vector):
		return Vector2D(self.x + vector.x, self.y + vector.y)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import unittest
from pyradium.VariableSubstitution import VariableSubstitutionContainer
from pyradium.Exceptions import InvalidFStringExpressionException

class VariableSubstitutionTests(unittest.TestCase):
	def test_merge_dicts_overwrite(self):
		dict_a = { "foo": "bar" }
		dict_b = { "foo": "moo" }
		self.assertEqual(VariableSubstitutionContainer.merge_dicts(dict_a, dict_b), { "foo": "moo" })

	def test_merge_dicts_merge(self):
		dict_a = { "foo": "bar" }
		dict_b = { "bar": "moo" }
		self.assertEqual(VariableSubstitutionContainer.merge_dicts(dict_a, dict_b), { "foo": "bar", "bar": "moo" })

	def test_merge_dicts_merge_recursive(self):
		dict_a = { "foo": { "x": 1 } }
		dict_b = { "foo": { "y": 2 } }
		self.assertEqual(VariableSubstitutionContainer.merge_dicts(dict_a, dict_b), { "foo": { "x": 1, "y": 2 } })

	def test_dict_lookup_no_substitution(self):
		sub_dict = VariableSubstitutionContainer({
			"foo": 3,
			"bar": "three",
			"sub": {
				"moo":	"koo",
			},
		})
		self.assertEqual(sub_dict["foo"], 3)
		self.assertEqual(sub_dict["bar"], "three")
		self.assertEqual(sub_dict["sub"]["moo"], "koo")

	def test_dict_lookup_simple_substitution(self):
		sub_dict = VariableSubstitutionContainer({
			"foo": 3,
			"bar": "{v['foo'] + 1:02d}",
		})
		self.assertEqual(sub_dict["foo"], 3)
		self.assertEqual(sub_dict["bar"], "04")

	def test_dict_lookup_nested_substitution1(self):
		sub_dict = VariableSubstitutionContainer({
			"char": "~",
			"foo": "{v['char'] * 3}",
			"bar": "{v['foo']:>5s} XYZ {v['foo']:<5s}",
		})
		self.assertEqual(sub_dict["foo"], "~~~")
		self.assertEqual(sub_dict["bar"], "  ~~~ XYZ ~~~  ")

	def test_dict_lookup_nested_substitution2(self):
		sub_dict = VariableSubstitutionContainer({
			"time": {
				"hrs": 23,
				"min": 55,
				"total_min": "{v['time']['hrs'] * 60 + v['time']['min']}",
				"remaining_min": "{int(v['time']['total_min']) - 1234}",
			},
			"remaining_time": "{int(v['time']['remaining_min']) // 60:02d}:{int(v['time']['remaining_min']) % 60:02d}",
		}, {
			"int":	int,
		})
		self.assertEqual(sub_dict["time"]["total_min"], "1435")
		self.assertEqual(sub_dict["time"]["remaining_min"], "201")
		self.assertEqual(sub_dict["remaining_time"], "03:21")

	def test_dict_infinite_loop(self):
		sub_dict = VariableSubstitutionContainer({
			"a": "{v['b']}",
			"b": "{v['a']}",
		})
		with self.assertRaises(InvalidFStringExpressionException):
			sub_dict["a"]

	def test_illegal_expression(self):
		sub_dict = VariableSubstitutionContainer({
			"a": "\"\"\"",
		})
		with self.assertRaises(InvalidFStringExpressionException):
			sub_dict["a"]

	def test_evaluate_all(self):
		sub_dict = VariableSubstitutionContainer({
			"time": {
				"hrs": 23,
				"min": 55,
				"total_min": "{v['time']['hrs'] * 60 + v['time']['min']}",
				"remaining_min": "{int(v['time']['total_min']) - 1234}",
			},
			"x": [ "{1+3}", "9", [ "{9-3}" ] ],
			"remaining_time": "{int(v['time']['remaining_min']) // 60:02d}:{int(v['time']['remaining_min']) % 60:02d}",
		}, {
			"int":	int,
		})
		self.assertEqual(sub_dict.evaluate_all(), {
			"time": {
				"hrs": 23,
				"min": 55,
				"total_min": "1435",
				"remaining_min": "201",
			},
			"x": [ "4", "9", [ "6" ] ],
			"remaining_time": "03:21"
		})



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from pyradium.Tools import XMLTools
from .SVGStyle import SVGStyle

class SVGText():
	def __init__(self, text_node):
		self._node = text_node
		self._node.setAttribute("xml:space", "preserve")
		self._textdata_node = None
		try:
			self._tspan_node = XMLTools.findfirst(self._node, "tspan")
			try:
				self._textdata_node = next(XMLTools.findall_text(self._tspan_node))
			except StopIteration:
				pass
		except StopIteration:
			self._tspan_node = self._node.appendChild(self._node.ownerDocument.createElement("tspan"))
		if self._textdata_node is None:
			self._textdata_node = self._tspan_node.appendChild(self._node.ownerDocument.createTextNode(""))
		self._style = SVGStyle.from_node(self._node, auto_sync = True).default_text()
		self._tspan_style = SVGStyle.from_node(self._tspan_node, auto_sync = True)

	@property
	def text(self):
		return self._textdata_node.wholeText

	@text.setter
	def text(self, value):
		self._textdata_node.replaceWholeText(value)

	@property
	def style(self):
		return self._style

	@property
	def tspan_style(self):
		return self._tspan_style

	@property
	def font(self):
		return self.tspan_style["font-family"]

	@font.setter
	def font(self, value):
		self.tspan_style["font-family"] = value



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import logging
from pyradium.DigitalTimingDiagram import DigitalTimingDiagram
from .BaseStandaloneCommand import BaseStandaloneCommand

_log = logging.getLogger(__spec__.name)

@BaseStandaloneCommand.register
class DigitalTimingDiagramRenderer(BaseStandaloneCommand):
	_NAME = "dtd"
	_DESCRIPTION = "Render digital timing diagram"

	@classmethod
	def _gen_parser(cls, parser):
		parser.add_argument("-H", "--low-high-lines", action = "store_true", help = "Render LOW/HIGH horizontal lines.")
		parser.add_argument("-x", "--xdiv-size", metavar = "pixels", type = float, default = 10, help = "Defines the size of the X division in pixels. Defaults to %(default).1f.")
		parser.add_argument("-y", "--height", metavar = "pixels", type = float, default = 30, help = "Defines the bar height in pixels. Defaults to %(default).1f.")
		parser.add_argument("-n", "--no-clock-ticks", action = "store_true", help = "Do not draw clock tick marks.")
		parser.add_argument("-f", "--force", action = "store_true", help = "Overwrite output file if it already exists.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("infile", help = "Input text representation of the digital timing diagram.")
		parser.add_argument("outfile_svg", help = "Output SVG file of the rendered digital timing diagram.")

	def run(self):
		if (not self._args.force) and os.path.exists(self._args.outfile_svg):
			print(f"Refusing to overwrite: {self._args.outfile_svg}")
			return 1

		with open(self._args.infile) as f:
			dtd_text = f.read()

		dtd = DigitalTimingDiagram(xdiv = self._args.xdiv_size, height = self._args.height, clock_ticks = not self._args.no_clock_ticks, low_high_lines = self._args.low_high_lines)
		dtd.parse_and_write(dtd_text)
		dtd.svg.writefile(self._args.outfile_svg)
		return 0



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import hashlib
from pyradium.xmlhooks.XMLHookRegistry import BaseHook, XMLHookRegistry
from pyradium.Exceptions import MalformedXMLInputException

@XMLHookRegistry.register_hook
class FileHook(BaseHook):
	_TAG_NAME = "file"

	@classmethod
	def handle(cls, rendered_presentation, node):
		if not node.hasAttribute("src"):
			raise MalformedXMLInputException("File node needs 'src' attribute set.")

		filename_key = rendered_presentation.renderer.presentation.meta.get("filename-key", "")
		src_filename = rendered_presentation.renderer.lookup_include(node.getAttribute("src"))
		basename = os.path.basename(src_filename)
		filename_hash = hashlib.md5((basename + filename_key).encode("utf-8")).hexdigest()
		if rendered_presentation.renderer.rendering_params.resource_uri != "":
			target_filename = f"{rendered_presentation.renderer.rendering_params.resource_uri}/incfiles/{filename_hash}/{basename}"
		else:
			target_filename = f"incfiles/{filename_hash}/{basename}"
		rendered_presentation.copy_abs_file(src_filename, target_filename)

		replacement_node = node.ownerDocument.createElement("a")
		replacement_node.setAttribute("href", target_filename)

		if len(node.childNodes) == 0:
			# Include the literal file basename as text
			replacement_node.appendChild(node.ownerDocument.createTextNode(basename))
		else:
			for child in list(node.childNodes):
				replacement_node.appendChild(child)

		return replacement_node



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2021-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyradium.Controller import BaseController

class TOCController(BaseController):
	@property
	def toc_items_per_slide(self):
		return self._options["toc_items_per_slide"]

	def render(self):
		toc = self.rendered_presentation.frozen_toc
		if toc is not None:
			start_at = self.slide.get_xml_slide_var("start_at")
			end_before = self.slide.get_xml_slide_var("end_before")
			toc_item_count = toc.count_toc_items(start_at = start_at, end_before = end_before)
			toc_slide_count = (toc_item_count + self.toc_items_per_slide - 1) // self.toc_items_per_slide

			additional_slide_var_list = [ ]
			for toc_slide_index in range(toc_slide_count):
				subset = list(toc.subset(start_at = start_at, end_before = end_before, max_items = self.toc_items_per_slide))
				start_at = subset[-1].index + 1
				additional_slide_vars = {
					"partial_toc":	toc.emit_commands(subset),
				}
				additional_slide_var_list.append(additional_slide_vars)
			yield from self.slide.emit_nocontent_slide(self.rendered_presentation, self.content_containers, additional_slide_var_list)



import setuptools

with open("README.md") as f:
	long_description = f.read()

setuptools.setup(
	name = "pyradium",
	packages = setuptools.find_packages(),
	version = "0.0.16",
	license = "gpl-3.0",
	description = "Creating HTML presentations with LaTeX-ish features from XML source",
	long_description = long_description,
	long_description_content_type = "text/markdown",
	author = "Johannes Bauer",
	author_email = "joe@johannes-bauer.com",
	url = "https://github.com/johndoe31415/pyradium",
	download_url = "https://github.com/johndoe31415/pyradium/archive/0.0.16.tar.gz",
	keywords = [ "latex", "presentation", "template", "html" ],
	install_requires = [
		"mako",
		"pygments",
		"requests",
		"lzstr>=0.0.3",
	],
	entry_points = {
		"console_scripts": [
			"pyradium = pyradium.__main__:main"
		]
	},
	include_package_data = True,
	classifiers = [
		"Development Status :: 5 - Production/Stable",
		"Intended Audience :: Developers",
		"License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3 :: Only",
		"Programming Language :: Python :: 3.5",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"Programming Language :: Python :: 3.8",
		"Programming Language :: Python :: 3.9",
	],
)
